#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Data manipulation abstractions.

Unless explicitly provided here, all `numpy.ndarray` methods and free functions
from `numpy` and `numpy.lib.recfunctions` should work properly with any type of
arrays (`Data` and non-`Data`).

"""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2009-2016, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import collections
import functools
import inspect
import itertools
import numpy
import numpy as np                      # required for some numpy docstrings
from numpy.lib import recfunctions
import re
import six
from six.moves import zip as _zip
import sys
import warnings
import weakref

from sciexp2.common import progress
import sciexp2.common.filter
from sciexp2.common import pp
from sciexp2.common import varref
import sciexp2.common.instance
from sciexp2.data import meta


"""
.. todo::

   redefine 'numpy.core.numeric.{save,savez,savetxt}' so that a message can be
   shown

.. todo::

   redefine 'numpy.core.numeric.load' so that a message can be shown

"""


######################################################################
# * Creation

def data_array(object, dims=None, *args, **kwargs):
    """Create a new array with dimension metadata.

    Parameters
    ----------
    object : array_like
        See `numpy.array`.
    dims : optional
        See argument `dims` in `Data` (default is ``object``).
    args, kwargs
        Extra arguments to `numpy.array`.

    See also
    --------
    Data, numpy.array

    """
    dtype = kwargs.pop("dtype", None)
    if dtype is None and len(args) > 0:
        try:
            dtype = numpy.dtype(args[0])
        except:
            pass
        if dtype is not None:
            args = args[1:]
    elif dtype is None and isinstance(object, numpy.ndarray):
        dtype = object.dtype
    if dtype is not None:
        kwargs["dtype"] = dtype

    shape = numpy.shape(object)
    if not isinstance(object, numpy.ndarray):
        # fixup shape if it's producing a structured array
        if isinstance(dtype, collections.Iterable):
            dtype = list(dtype)
            if dtype is not None and len(shape) > 1 and \
               shape[-1] == len(dtype):
                shape = shape[:-1]

    if dims is None and isinstance(object, Data):
        dims = object

    res = Data(shape, dims, *args, **kwargs)
    res.view(np.ndarray)[:] = object
    return res


def data_frombuffer(buffer, dims, *args, **kwargs):
    """Create an array with dimension metadata from a buffer.

    Parameters
    ----------
    buffer
        See argument `buffer` in `numpy.frombuffer`.
    dims
        See argument `dims` in `Data`.
    args, kwargs
        Arguments to `numpy.frombuffer`.

    See also
    --------
    numpy.frombuffer

    """
    arr = numpy.frombuffer(buffer, *args, **kwargs)
    return Data(shape=arr.shape, dims=dims, dtype=arr.dtype, buffer=arr)


def data_memmap(filename, dims, *args, **kwargs):
    """Create an array with dimension metadata from a buffer.

    Parameters
    ----------
    filename
        See argument `filename` in `numpy.memmap`.
    dims
        See argument `dims` in `Data`.
    args, kwargs
        Arguments to `numpy.memmap`.

    See also
    --------
    numpy.memmap

    """
    arr = numpy.memmap(filename, *args, **kwargs)
    res = Data(shape=arr.shape, dims=dims, dtype=arr.dtype, buffer=arr)
    # attributes
    res._memmap_array = arr
    res.filename = arr.filename
    res.offset = arr.offset
    res.mode = arr.mode
    # methods
    res.flush = arr.flush
    return res


######################################################################
# * Data

##################################################
# ** ufunc helpers

def _wrap_npfunc(npfunc, wrapper):
    """Wrap numpy ufuncs."""

    wrapper.func_name = npfunc.__name__
    wrapper.func_doc = npfunc.__doc__
    wrapper.__name__ = npfunc.__name__
    wrapper.__doc__ = npfunc.__doc__
    return wrapper


def _get_axis(axis, args):
    if isinstance(args, dict):
        datas = [val
                 for key, val in args.items()
                 if key != "out" and isinstance(val, Data)]
    else:
        datas = [val
                 for val in args
                 if isinstance(val, Data)]

    if len(datas) == 0:
        return None, axis
    elif len(datas) == 1:
        arg = datas[0]
        if isinstance(axis, tuple):
            new_axis = []
            for a in axis:
                if isinstance(a, six.string_types):
                    a = arg.dims[a].index
                new_axis.append(a)
            return arg, tuple(new_axis)
        elif axis is None:
            return arg, axis
        else:
            return arg, arg.dims[axis].index
    else:
        res_data = None
        res_axis = None
        for data in datas:
            _, data_axis = _get_axis(axis, [data])
            if res_axis is None:
                res_data = data
                res_axis = data_axis
            elif res_axis != data_axis:
                raise IndexError(
                    "axis translates to different dimensions on each argument")
        return res_data, res_axis


def wrap_reduce(func):
    """Wrap reductions to keep dimension metadata consistent.

    Due to numpy internals, this wrapper might raise a `ShapeError`
    exception. If so, the manual equivalent of:

    >>> f_wrap = wrap_reduce(func)  # doctest: +SKIP
    >>> d = f_wrap(data, axis=a)    # doctest: +SKIP

    is:

    >>> d = func(data, axis=a)          # doctest: +SKIP
    >>> d_dims = DataDims(data.dims)    # doctest: +SKIP
    >>> del d_dims[a]                   # doctest: +SKIP
    >>> d = Data(d, dims=d_dims)        # doctest: +SKIP

    """
    func_proto = func
    if not inspect.isfunction(func):
        func_proto = getattr(numpy, func.__name__)
    func_argspec = inspect.getargspec(func_proto)
    assert "axis" in func_argspec.args
    func_axis_index = func_argspec.args.index("axis")
    if "out" in func_argspec.args:
        func_out_index = func_argspec.args.index("out")
    else:
        func_out_index = None

    def wrapper(*args, **kwargs):
        callargs = inspect.getcallargs(func_proto, *args, **kwargs)
        keepdims = callargs["keepdims"]
        if len(args) <= func_axis_index:
            axis = callargs["axis"]
        else:
            axis = args[func_axis_index]
        if func_out_index is not None and len(args) <= func_out_index:
            out = callargs["out"]
        else:
            out = None

        arg, axis = _get_axis(axis, callargs)

        if len(args) <= func_axis_index:
            kwargs["axis"] = axis
        else:
            args = list(args)
            args[func_axis_index] = axis

        res = func(*args, **kwargs)

        def get_dims(dims, axis, keepdims):
            assert keepdims or axis is not None
            if axis is None:
                axis = tuple(range(len(dims)))
            if isinstance(axis, tuple):
                if keepdims:
                    new_dims = list(dims)
                    for a in axis:
                        new_dims[a] = meta.Dim(
                            new_dims[a]._expression, [None])
                    res = DataDims(new_dims)
                else:
                    res = DataDims(dims)
                    del res[list(axis)]
            else:
                if keepdims:
                    new_dims = list(dims)
                    new_dims[axis] = meta.Dim(
                        new_dims[axis]._expression, [None])
                    res = DataDims(new_dims)
                else:
                    res = DataDims(dims)
                    del res[axis]
            return res

        if isinstance(res, Data) and not isinstance(out, Data):
            if res.ndim == 0:
                return res.dtype.type(res)

            if axis is not None:
                if not _check_shape(res):
                    assert arg
                    res.dims[:] = get_dims(arg.dims, axis, keepdims)
            else:
                # had zero-dimensional result
                pass
            _ensure_shape(res)
        elif (axis is not None or keepdims) and \
             out is None and numpy.shape(res) != tuple():
            res = data_array(res, dims=get_dims(arg.dims, axis, keepdims))

        return res

    return _wrap_npfunc(func, wrapper)


def wrap_accumulate(func):
    """Wrap accumulations to keep dimension metadata consistent."""
    func_proto = func
    if not inspect.isfunction(func):
        func_proto = getattr(numpy, func.__name__)
    func_argspec = inspect.getargspec(func_proto)
    assert "axis" in func_argspec.args
    func_axis_index = func_argspec.args.index("axis")

    def wrapper(*args, **kwargs):
        callargs = inspect.getcallargs(func_proto, *args, **kwargs)
        if len(args) <= func_axis_index:
            axis = callargs["axis"]
        else:
            axis = args[func_axis_index]

        arg, axis = _get_axis(axis, callargs)

        if len(args) <= func_axis_index:
            kwargs["axis"] = axis
        else:
            args = list(args)
            args[func_axis_index] = axis

        res = func(*args, **kwargs)

        if isinstance(res, Data):
            if axis is None:
                assert arg is not None and res.size == arg.size
                res = res.view(numpy.ndarray)
            else:
                _ensure_shape(res)

        return res

    return _wrap_npfunc(func, wrapper)


##################################################
# ** misc helpers

class ShapeError (Exception):
    """Mismatching array and metadata shape."""
    def __init__(self, data_shape, dims_shape):
        self.data_shape = data_shape
        self.dims_shape = dims_shape

    def __str__(self):
        return "Data object shape '%s' and dimension metadata "\
            "'%s' are not consistent" % (self.data_shape, self.dims_shape)


def _check_shape(arr, dims=None):
    """Check if the array and the dimension metadata have the same size."""
    if dims is None:
        dims = arr.dims

    if arr.ndim != len(dims):
        return False
    for nd in range(arr.ndim):
        if arr.shape[nd] != len(dims[nd]):
            return False
    return True


def _ensure_shape(arr):
    """Ensure the array and the dimension metadata have the same size."""
    dims = arr.dims
    if not _check_shape(arr, dims):
        dims_shape = tuple([len(d) for d in dims])
        raise ShapeError(arr.shape, dims_shape)


class ReshapeError (Exception):
    """Error during `Data.reshape`."""

    def __init__(self, expression, elem1, elem2, variables):
        super(ReshapeError, self).__init__(expression,
                                           dict(elem1), dict(elem2),
                                           variables)

    def __str__(self):
        msg = "Ambiguous dimensions; unused variables: "
        msg += ", ".join(self.args[3])
        msg += "\nConflicting source elements with expression: %s\n" %\
               self.args[0]
        msg += "   %s\n" % (self.args[1])
        msg += "   %s" % (self.args[2])
        return msg


def _reshape_parse_fields_to_vars(f2v, data):
    res = []
    for f2v_re in f2v:
        match = False
        f2v_cre = re.compile(f2v_re)
        for data_field in data.dtype.names:
            if f2v_cre.match(data_field) is not None:
                match = True
                if data_field in res:
                    continue
                if data_field in data.dims[0].variables:
                    raise ValueError("Field in `fields_to_vars` already "
                                     "exists as a variable: %s" % data_field)
                res.append(data_field)
        if not match:
            raise ValueError("Value in `fields_to_vars` does not match "
                             "any field: %s" % f2v_re)
    if len(data.dtype) > 0 and set(res) == set(data.dtype.names):
        raise ValueError("Cannot move all fields into variables with "
                         "`fields_to_vars`")
    return res


def _reshape_parse_vars_to_fields(v2f, data):
    res = []
    for v2f_re in v2f:
        match = False
        v2f_cre = re.compile(v2f_re)
        for data_var in data.dims[0].variables:
            if v2f_cre.match(data_var) is not None:
                match = True
                if data_var in res:
                    continue
                if data_var in data.dtype.names:
                    raise ValueError("Variable in `vars_to_fields` already "
                                     "exists as a field: %s" % data_var)
                res.append(data_var)
        if not match:
            raise ValueError("Value in `vars_to_fields` does not match "
                             "any variable: %s" % v2f_re)
    if len(data.dtype) > 0 and set(res) == set(data.dims[0].variables):
        raise ValueError("Cannot move all variables into fields with "
                         "`vars_to_fields`")
    return res


def _reshape_compute_data(f2v, v2f, data):
    if len(f2v) != 0 and len(f2v) < len(data.dtype):
        data = recfunctions.drop_fields(data.view(numpy.ndarray), f2v)
    if len(v2f) != 0:
        new_data = [numpy.array([elem[var] for elem in data.dims[0].instances])
                    for var in v2f]
        if len(f2v) == len(data.dtype):
            data = append_fields(new_data[0], names=v2f[1:], data=new_data[1:])
        else:
            data = append_fields(data, names=v2f, data=new_data)
    return data


def _reshape_compute_dimension(f2v, v2f, data):
    if len(f2v) == 0 and len(v2f) == 0:
        return data.dims[0]
    expression = data.dims[0].expression
    if len(v2f) > 0:
        variables = [var for var in data.dims[0].variables
                     if var not in v2f]
        expression = "@%s@" % "@-@".join(variables)
    if len(f2v) > 0:
        expression += "-@%s@" % "@-@".join(f2v)

    contents = []
    with progress.progressable_simple(
            data.dims[0].instances, None,
            msg="recomputing dimension contents...") as old_dimension:
        for elem in old_dimension:
            elem_idx = data.dims[0].indexes[elem]
            elem = elem.copy()
            for field in f2v:
                elem[field] = data[elem_idx][field]
            contents.append(elem)
    return meta.Dim(expression, contents)


def _reshape_parse_filling_values(filling_values, dtype):
    if isinstance(filling_values, dict):
        if len(dtype) == 0:
            raise ValueError("Invalid `filling_values`, array has no "
                             "fields")
        for field in filling_values:
            if field not in dtype.names:
                raise ValueError("Unknown field name %r in "
                                 "'filling_values'" % field)
    elif len(dtype) > 0 and filling_values is not None:
        filling_values = dict([(name, filling_values)
                               for name in dtype.names])
    return _reshape_filling_values(filling_values, dtype)


def _reshape_parse_ignore(ignore, dim_variables):
    ignore_rest = False
    if ignore is True:
        ignore_rest = True
        ignore = set()
    else:
        if not isinstance(ignore, list) and not isinstance(ignore, set):
            raise TypeError("Unknown `ignore` value: %s" % ignore)
        ignore = set(ignore)
        source_variables = set(dim_variables)
        if len(ignore - source_variables) > 0:
            raise ValueError("Invalid variables in `ignore`: " +
                             ", ".join(ignore - source_variables))
    return ignore, ignore_rest


def _reshape_filling_values(filling_values, dtype):
    if len(dtype) == 0:
        fdtype = numpy.dtype([("f0", dtype)])
    fdtype = numpy.dtype(dtype.descr + [("__flast__", int)])

    contents = "," * (len(fdtype) - 1)
    contents += "1"
    values = numpy.genfromtxt(six.StringIO(contents),
                              delimiter=",",
                              dtype=fdtype)

    if len(dtype) == 0:
        if filling_values is not None:
            try:
                values["f0"] = filling_values
            except:
                raise ValueError("Incompatible `filling_values` type")
        return values["f0"]
    else:
        if filling_values is not None:
            for k, v in six.iteritems(filling_values):
                try:
                    values[k] = v
                except:
                    raise ValueError("Incompatible `filling_values` type for "
                                     "field %r" % k)
        return tuple(values.tolist()[:-1])


##################################################
# ** main code

def _indexes_normalize(index, ndim, identify=False):
    """
    Examples
    --------

    >>> norm = _indexes_normalize
    >>> norm(1, 3)
    [1, slice(None, None, None), slice(None, None, None)]

    >>> norm((1, Ellipsis), 3)
    [1, slice(None, None, None), slice(None, None, None)]
    >>> norm((1, np.newaxis), 3)
    [1, None, slice(None, None, None), slice(None, None, None)]
    >>> norm((np.newaxis, 1), 3)
    [None, 1, slice(None, None, None), slice(None, None, None)]
    >>> norm((np.newaxis, 1, Ellipsis), 3)
    [None, 1, slice(None, None, None), slice(None, None, None)]

    >>> norm((Ellipsis, 1), 3)
    [slice(None, None, None), slice(None, None, None), 1]
    >>> norm((Ellipsis, 1, np.newaxis), 3)
    [slice(None, None, None), slice(None, None, None), 1, None]
    >>> norm((np.newaxis, Ellipsis, 1), 3)
    [None, slice(None, None, None), slice(None, None, None), 1]
    >>> norm((Ellipsis, np.newaxis, 1), 3)
    [slice(None, None, None), slice(None, None, None), None, 1]

    """

    # http://docs.scipy.org/doc/numpy/reference/arrays.indexing.html#basic-slicing
    #
    # "In order to remain backward compatible with a common usage in
    # Numeric, basic slicing is also initiated if the selection object is
    # any sequence (such as a list) containing slice objects, the Ellipsis
    # object, or the newaxis object, but no integer arrays or other
    # embedded sequences."
    def is_numeric_basic_slicing(indexes):
        return \
            isinstance(indexes, collections.Iterable) and \
            all(isinstance(idx, slice) or
                idx is Ellipsis or
                idx is numpy.newaxis
                for idx in indexes)

    # normalize format to a sequence of indexes
    indexes = index
    if not isinstance(indexes, tuple) and \
       not is_numeric_basic_slicing(indexes):
        indexes = [(True, indexes)]
    else:
        indexes = list((True, idx) for idx in indexes)

    # expand Ellipsis into appropriate number of empty slices
    num_newaxis = 0
    for _, i in indexes:
        if i is np.newaxis:
            num_newaxis += 1
    first_ellipsis = -1
    for idx, (_, elem) in enumerate(indexes):
        if elem is Ellipsis:
            first_ellipsis = idx
            break
    if first_ellipsis >= 0:
        indexes = [(exist, idx) if idx is not Ellipsis else (True, slice(None))
                   for exist, idx in indexes]
        span = ndim - len(indexes) + 1 + num_newaxis
        if span < 0:
            raise IndexError("too many indices")
        indexes = (indexes[:first_ellipsis] +
                   ([(False, slice(None))] * span) +
                   indexes[first_ellipsis+1:])
    else:
        span = ndim - len(indexes) + num_newaxis
        if span < 0:
            raise IndexError("too many indices")

    # add empty slices for trailing dimensions
    indexes += [(False, slice(None))] * (ndim - len(indexes) + num_newaxis)

    if not identify:
        indexes = [idx for _, idx, in indexes]
    return indexes


def _indexes_trim_suffix(indexes):
    """Trim trailing empty slices."""
    indexes = list(indexes)
    while len(indexes) > 1:
        last = indexes[-1]
        if isinstance(last, slice) and \
           last.start == None and \
           last.stop == None and \
           last.step == None:
            indexes.pop()
        else:
            break
    return indexes


class Data (numpy.ndarray, pp.Pretty):
    """Data(shape, dims, *args, **kwargs)

    N-dimensional array with dimension metadata.

    Parameters
    ----------
    shape : tuple(int)
        See `numpy.ndarray`.
    dims : `DataDims` or list, optional
        `meta.Dim` metadata.
    args, kwargs
        Extra arguments to `numpy.ndarray`.

    Notes
    -----
    The `dims` argument can be a `DataDims` instance or a sequence of valid
    arguments to its constructor. If `dims` is a list, but with less elements
    than dimensions in the array, additional empty `meta.Dim` objects will be
    added for the remaining dimensions.

    Functions with an `axis` argument can identify the target axis (`meta.Dim`)
    using the same semantics for indexing on the `dims` attribute.

    Element-wise comparison ignores the array metadata.

    See also
    --------
    data_array, numpy.ndarray, sciexp2.common.filter, reshape, meta.Dim,
    meta.Dim.sort

    Examples
    --------
    Data arrays can be accessed like regular arrays:

    >>> d = data_array([[(1, 2), (3, 4), (5, 6)],
    ...                 [(7, 8), (9, 10), (11, 12)]],
    ...                dtype = [("time", float), ("energy", float)],
    ...                dims=[("@cores@",
    ...                       ["1", "2"]),
    ...                      ("@entries@-@entry_size@",
    ...                       ["1-8", "2-8", "2-16"])])
    >>> d[0,0]["time"]
    1.0
    >>> d[0,0:2]["time"]
    data_array([1.0, 3.0], dims=[('@entries@-@entry_size@', ['1-8', '2-8'])])

    But indexing with strings will use the dimension metadata instead:

    >>> d["1","1-8"]["time"] == d[0,0]["time"]
    True
    >>> d.dims["cores"].indexes["1"] == 0 and d.dims["entries"].indexes["1-8"] == 0
    True
    >>> (d["1","1-8":"2-16"]["time"] == d[0,0:2]["time"]).all()
    True
    >>> d.dims["entries"].indexes["2-16"] == 2
    True

    Additionally, you can use filters to select specific entries, by using the
    stepping indexing argument in the slicing syntax:

    >>> d[:,::2]
    data_array([[(1.0, 2.0), (5.0, 6.0)], [(7.0, 8.0), (11.0, 12.0)]],
               dtype=[('time', '<f8'), ('energy', '<f8')],
               dims=[('@cores@', ['1', '2']),
                     ('@entries@-@entry_size@', ['1-8', '2-16'])])
    >>> d[:,::"entry_size == 8"]
    data_array([[(1.0, 2.0), (3.0, 4.0)], [(7.0, 8.0), (9.0, 10.0)]],
               dtype=[('time', '<f8'), ('energy', '<f8')],
               dims=[('@cores@', ['1', '2']),
                     ('@entries@-@entry_size@', ['1-8', '2-8'])])
    >>> d[:,::"entries == 1 && entry_size == 8"]
    data_array([[(1.0, 2.0)], [(7.0, 8.0)]],
               dtype=[('time', '<f8'), ('energy', '<f8')],
               dims=[('@cores@', ['1', '2']), ('@entries@-@entry_size@', ['1-8'])])
    >>> d[:,"entries == 1 && entry_size == 8"]
    data_array([(1.0, 2.0), (7.0, 8.0)],
               dtype=[('time', '<f8'), ('energy', '<f8')],
               dims=[('@cores@', ['1', '2'])])

    """

    _XLATOR = varref.DictXlator("@")
    _EXPANDER = sciexp2.common.instance.Expander("@")

    ##########################################################################
    # creation
    ##########################################################################

    def __new__(cls, shape, dims=None, *args, **kwargs):
        res = numpy.ndarray.__new__(cls, shape, *args, **kwargs)

        if dims is None:
            dims = []
        elif isinstance(dims, Data):
            dims = list(dims.dims)
        dims = list(dims)

        res._dims = DataDims([(None, [None] * d)
                              for d in list(res.shape)])
        res._dims._array = res
        res.dims[:len(dims)] = dims

        return res

    def __array_finalize__(self, obj):
        if obj is None:                    # explicit constructor
            pass
        elif isinstance(obj, Data):
            # just copy and hope the caller will patch it up
            self._dims = obj._dims

        else:                           # view casting into Data
            raise TypeError("cannot cast a view into a Data object")

    def __array_wrap__(self, obj, context=None):
        result = obj
        if result.ndim == 0:
            return result.item()
        if context is not None and len(context[1]) > 1:  # binary ufunc
            args = [arg
                    for arg in context[1]
                    if numpy.shape(arg) == result.shape]
            assert len(args) > 0

            def find_match():
                for arg in args:
                    # TODO: does not align metadata if multiple args match
                    if isinstance(arg, Data):
                        return arg
                return None

            match = find_match()
            if match is not None:
                if not isinstance(result, Data):
                    result = result.view(type(self))
                result._dims = match._dims.copy()
                result._dims._array = result
            else:
                if isinstance(result, Data):
                    result._dims = DataDims([(None, [None] * d)
                                             for d in list(result.shape)])
                    result._dims._array = result

        if isinstance(result, Data):
            _ensure_shape(result)
        return result

    def copy(self, order="C"):
        return data_array(self, order=order)

    copy.__doc__ = numpy.ndarray.copy.__doc__

    def _repr_pretty_(self, p, cycle):
        with self.pformat(p, cycle, name="data_array"):
            p.pretty(self.tolist())
            nd_repr = repr(self.flat[0:1].view(numpy.ndarray))
            if re.search(r"dtype=.*\)\s*$", nd_repr, re.MULTILINE) is not None:
                p.text(",")
                p.breakable()
                p.text("dtype=")
                with p.group(len("dtype="), "", ""):
                    descr = self.dtype.descr
                    if len(descr) == 1 and descr[0][0] == "":
                        p.pretty(descr[0][1])
                    else:
                        p.pretty(descr)

            p.text(",")
            p.breakable()
            p.text("dims=")
            with p.group(len("dims="), "", ""):
                dims = []
                for dim in self.dims:
                    if dim._have_expression:
                        dims += [(dim.expression, list(dim))]
                    elif list(dim._group.variables()) == [meta.Dim._DEFAULT_VAR]\
                         and len(dim._group[meta.Dim._DEFAULT_VAR]) == 0:
                        dims += [None]
                    else:
                        dims += [(None, list(dim))]
                p.pretty(dims)

    def __repr__(self):
        return pp.Pretty.__repr__(self)

    def reshape(self, *shape, **kwargs):
        """Return a copy of the array with a different shape.

        Compared to `numpy.reshape`, `Data.reshape` indicates the new shape by
        establishing which dimension metadata variables are to describe each
        dimension.


        Parameters
        ----------
        shape
            List describing each dimension of the resulting array.
        ignore : list of str, optional
            Variables that will not be used to describe dimensions. Defaults to
            an empty list, which will implicitly add a last dimension with all
            the variables not referenced in `shape`.
        filling_values : value or dict, optional
            Default values to fill-in fields previously unexisting elements
            created by reshape.
        fields_to_vars : list of str, optional
            Names of fields in the array that will be instead used as dimension
            variables (field values will be used as dimension tick values). You
            can also use regular expressions to select the matching fields.
        vars_to_fields : list of str, optional
            Names of dimension variables that will be instead used as fields in
            the result (variable values will be used as field values). You can
            also use regular expressions to select the matching variables.


        Returns
        -------
        Data
            Copy of the array with the new shape.


        Raises
        ------
        ReshapeError
            The specified `shape` is ignoring variables that have multiple
            values for a single destination element.


        Notes
        -----
        Each element in `shape` can have one of the following formats:

        * ``tuple('expression', 'filter')``
          Dimension is described by the given expression, but only taking
          elements conforming to the given filter.

        * ``'expression'``
          Shorthand to use with empty filters.

        * ``tuple(list('variable', ...), 'filter')``
          Alternative to the first format, where the list of variables is used
          to automatically build an expression.

        * ``list('variable', ...)``
          Shorthand to use with empty filters.

        Reshaping can help in working with data more efficiently, since direct
        indexing is faster than slicing filters (e.g., ``d[::"a==2"]``); but it
        can also be used to have arbitrary expressions in the dimension
        indices.

        Reshaping an array can result in more elements than the original. The
        newly-added elements are filled-in with default values, which can be
        controlled through `filling_values`. A simple value will set all fields
        to that. Using a dictionary will map field names to their default
        value.

        Examples
        --------
        Values for artificially-added elements can be controlled:

        >>> d = data_array([1, 2, 3, 4, 5],
        ...                dtype=int,
        ...                dims=[('@a@-@b@',
        ...                       ['1-1', '1-2', '1-3', '2-1', '2-2'])])
        >>> d.reshape(["a"])
        data_array([[1, 2, 3], [4, 5, -1]],
                   dims=[('@a@', ['1', '2']), ('@b@', ['1', '2', '3'])])
        >>> d.reshape(["a"], filling_values=0)
        data_array([[1, 2, 3], [4, 5, 0]],
                   dims=[('@a@', ['1', '2']), ('@b@', ['1', '2', '3'])])

        And arbitrary expressions can help later in generating axis tick names
        for plots:

        >>> d.reshape("a: @a@")
        data_array([[1, 2, 3], [4, 5, -1]],
                   dims=[('a: @a@', ['a: 1', 'a: 2']), ('@b@', ['1', '2', '3'])])

        But not all variables can be ignored. For example, ``b`` cannot be
        ignored because indices ``1-1`` and ``1-2`` have different values for
        it, unless we filter-out the offending values for ``b``:

        >>> d.reshape(["a"], ignore = True)
        Traceback (most recent call last):
            ...
        ReshapeError: Ambiguous dimensions; unused variables: b
        Conflicting source elements with expression: @a@-@b@
           {'a': 1, 'b': 1}
           {'a': 1, 'b': 2}
        >>> d.reshape((["a"], "b==2"), ignore = True)
        data_array([2, 5], dims=[('@a@', ['1', '2'])])


        See also
        --------
        numpy.reshape

        """
        if len(self.shape) != 1:
            raise ValueError("Can only reshape 1-dimensional Data objects")

        f2v = _reshape_parse_fields_to_vars(
            kwargs.pop("fields_to_vars", []),
            self)
        if "vars_to_fields" in kwargs and len(self.dtype) == 0:
            raise ValueError(
                "Cannot use 'vars_to_fields' on arrays with no fields")
        v2f = _reshape_parse_vars_to_fields(
            kwargs.pop("vars_to_fields", []),
            self)

        data = _reshape_compute_data(f2v, v2f, self)
        dimension = _reshape_compute_dimension(f2v, v2f, self)

        filling_values = _reshape_parse_filling_values(
            kwargs.pop("filling_values", None),
            data.dtype)

        ignore, ignore_rest = _reshape_parse_ignore(
            kwargs.pop("ignore", set()),
            dimension.variables)

        if len(kwargs) != 0:
            raise TypeError("Unknown arguments: " + ", ".join(kwargs.keys()))

        ##################################################
        # 0th phase
        #
        # Parse dimension descriptors.

        def parse_dimension(descriptor, prev_filter, variables, used_vars):
            """Parse and validate dimension descriptor."""
            # parsing
            res_filter = sciexp2.common.filter.Filter()
            if isinstance(descriptor, tuple):
                if len(descriptor) != 2:
                    raise ValueError("Invalid dimension descriptor: " +
                                     str(descriptor))
                res_expression = descriptor[0]
                res_filter &= descriptor[1]
            elif isinstance(descriptor, six.string_types) or\
                 isinstance(descriptor, list):
                res_expression = descriptor
            else:
                raise ValueError("Invalid dimension descriptor: " +
                                 str(descriptor))

            if isinstance(res_expression, list):
                res_expression = "@" + "@-@".join(res_expression) + "@"

            res_expression = meta.DimExpression(
                res_expression,
                varre=self.dims[0]._expression.varre,
                validate=variables)

            # validation
            res_variables = res_expression.variables
            for var in res_variables:
                if var in used_vars:
                    raise ValueError("Already used variable: %s" % var)
            # validates against the whole source, so that filters can
            # disambiguate elements when ignoring variables
            res_filter.validate(variables)

            used_vars += res_variables
            res_filter = prev_filter & res_filter
            return (res_expression, res_filter, used_vars)

        res_dim_expressions = []
        res_dim_filters = []
        res_dim_used_vars = []
        prev_filter = sciexp2.common.filter.Filter()

        # declared dimensions
        for dim_idx, descriptor in enumerate(shape):
            try:
                expression, filter_, res_dim_used_vars = parse_dimension(
                    descriptor, prev_filter, dimension.variables,
                    res_dim_used_vars)
            except (meta.DimExpressionError, ValueError,
                    sciexp2.common.filter.ParseError,
                    sciexp2.common.filter.FilterUnallowedError) as e:
                _, _, exc_traceback = sys.exc_info()
                new_e = ValueError(
                    ("While parsing descriptor for dimension %d:\n" % dim_idx) +
                    ("%s: %s" % (e.__class__.__name__, e)))
                six.reraise(new_e.__class__, new_e, exc_traceback)
            res_dim_expressions.append(expression)
            res_dim_filters.append(filter_)
            prev_filter = filter_

        # last dimension
        unused_variables = [
            variable for variable in dimension.variables
            if variable not in res_dim_used_vars and
            variable not in ignore]
        if not ignore_rest and len(unused_variables) > 0:
            try:
                expression, filter_, res_dim_used_vars = parse_dimension(
                    unused_variables, prev_filter, dimension.variables,
                    res_dim_used_vars)
            except (meta.DimExpressionError,
                    sciexp2.common.filter.ParseError,
                    sciexp2.common.filter.FilterUnallowedError) as e:
                _, _, exc_traceback = sys.exc_info()
                new_e = ValueError(
                    "While parsing descriptor for last dimension:\n"
                    "%s: %s" % (e.__class__.__name__, e))
                six.reraise(new_e.__class__, new_e, exc_traceback)
            res_dim_expressions.append(expression)
            res_dim_filters.append(filter_)

        ##################################################
        # 1st phase
        #
        # Build Dim objects for the resulting structure, while caching in
        # `res_maps` the mapping from the position of each element in `data` to
        # the position in the resulting array.

        res_dims = []         # list of Dim objects for `res`
        res_maps = []         # per-Dim maps from `data` to `res` indices

        for dim_idx, (expression, filter_) in enumerate(
                _zip(res_dim_expressions, res_dim_filters)):
            # NOTE: use internal attribute for speed
            self_dim = dimension._group

            with progress.progressable_simple(
                    self_dim, None,
                    msg="calculating dimension %d..." % dim_idx) as self_dim_progr:
                expansions = self._EXPANDER.expand(expression.expression,
                                                   self_dim_progr,
                                                   filter_=filter_,
                                                   recursive=False,
                                                   grouped=False)
            dim_map = {}

            # iterable with length
            class DimBuild (object):

                def __len__(self):
                    return len(expansions)

                def __iter__(self):
                    for dim_idx, expansion in enumerate(
                            six.itervalues(expansions)):
                        with progress.progressable_simple(
                                expansion.instances, None,
                                msg="building dimension %d..." % dim_idx) as instances:
                            instances = iter(instances)
                            try:
                                first = next(instances)
                            except StopIteration:
                                return
                            first_idx = dimension.indexes[first]
                            dim_map[first_idx] = dim_idx
                            for instance in instances:
                                instance_idx = dimension.indexes[instance]
                                dim_map[instance_idx] = dim_idx
                            yield first

            dim = meta.Dim(expression, DimBuild())

            res_dims.append(dim)
            res_maps.append(dim_map)

        ##################################################
        # 2nd phase
        #
        # Migrate data from `data` to the resulting array, using the
        # pre-computed index maps in `res_maps`.

        res_shape = tuple(len(d) for d in res_dims)

        # short-cut when shape does not change
        if list(res_shape) == list(data.shape):
            return data_array(data, dims=res_dims)

        res = Data(res_shape,
                   dims=res_dims,
                   dtype=data.dtype)

        with progress.get(res.size + 2, msg="reshaping contents...") as progr:
            progr()
            # set all default values at once for items artificially added to
            # maintain shape homogeneity
            res[:] = filling_values
            progr()
            res._fill_reshape(data, res_maps, progr)

        _ensure_shape(res)

        return res

    def _fill_reshape(self, source, self_maps, progr):
        # not space-efficient, but really fast and (hopefully) small enough
        res_set = numpy.ndarray(self.shape, dtype=int)
        res_set[:] = -1

        for source_pos in range(source.size):
            progr()
            try:
                self_pos = tuple([self_map[source_pos]
                                  for self_map in self_maps])
            except KeyError:
                # source position ignored by dimension filter
                continue
            if res_set[self_pos] >= 0:
                source1 = source.dims[0].instances[res_set[self_pos]]
                source2 = source.dims[0].instances[source_pos]
                unused = dict(source1)
                for k, v in source2:
                    if k in unused and unused[k] == v:
                        del unused[k]
                raise ReshapeError(source.dims[0].expression,
                                   source1, source2, list(unused.keys()))
            res_set[self_pos] = source_pos
            source_value = numpy.ndarray.__getitem__(
                source.view(np.ndarray), source_pos)
            numpy.ndarray.__setitem__(
                self.view(np.ndarray), self_pos, source_value)

    def _sort(self, dim, old_indexes):
        idx = self.indexer.dims[dim.index]
        self[idx[:]] = self[idx[old_indexes]]

    def __reduce__(self):
        count = sum([len(dim) for dim in self.dims])

        start = progress.get_pickle(count)
        stop = start.get_stop()

        object_state = list(numpy.ndarray.__reduce__(self))
        subclass_state = self._dims
        object_state[2] = (start, object_state[2], subclass_state, stop)
        return tuple(object_state)

    def __setstate__(self, state):
        _, nd_state, own_state, _ = state
        numpy.ndarray.__setstate__(self, nd_state)
        self._dims = own_state
        self._dims._array = self

    ##########################################################################
    # Dimension metadata accessors / manipulators
    ##########################################################################

    @property
    def dims(self):
        """`DataDims` with the dimension metadata of this array."""
        return self._dims

    @property
    def indexer(self):
        """`DataIndexer` for this array."""
        return DataIndexer(self)

    @property
    def idata(self):
        """`DataIData` for this array."""
        return DataIData(self)

    ##########################################################################
    # Item accessors / manipulators
    ##########################################################################

    def _get_indexes(self, index, need_dims):
        if isinstance(index, DataIndexer):
            index = index._get_indexes()
        indexes = _indexes_normalize(index, self.ndim)

        def calculate_fallback():
            state["copy"] = True
            state["advanced"] += 1
            if state["advanced"] > 1:
                # do not try to calculate metadata
                state["need_dims"] = False
                state["new_dims"] = None

        def reraise(exception):
            _, _, exc_traceback = sys.exc_info()
            new_e = IndexError(
                "error indexing dimension %d"
                "\n  [was] "
                "%s: %s" % (cur_dim,
                            e.__class__.__name__, str(e)))
            six.reraise(new_e.__class__, new_e, exc_traceback)

        state = dict(need_dims=need_dims,
                     copy=False,
                     advanced=0,
                     new_dims=[])
        cur_dim = 0
        new_indexes = [([], [])]

        for index in indexes:
            if index is numpy.newaxis:
                state["new_dims"].append(meta.Dim(None, [None]))
                new_indexes[-1][0].append(slice(None))
                new_indexes[-1][1].append(index)
            else:
                if isinstance(index, sciexp2.common.filter.PFilter):
                    index_shape = tuple()
                else:
                    index_shape = numpy.shape(index)

                append_dims = False

                if len(index_shape) > 0:  # advanced indexing (numpy)
                    append_dims = True
                    calculate_fallback()
                    if len(index_shape) > 1:
                        calculate_fallback()

                if state["need_dims"]:
                    # NOTE: get both in a single call
                    try:
                        new_dim, new_index = self.dims[cur_dim]._view(index)
                    except Exception as e:
                        reraise(e)
                else:
                    new_dim = None
                    try:
                        new_index = self.dims[cur_dim]._get_indexes(index)
                    except Exception as e:
                        reraise(e)

                is_filter = False
                new_index_shape = numpy.shape(new_index)

                if len(index_shape) == 0:
                    if len(new_index_shape) > 0:
                        # complex filter: must be applied on a separate access
                        # to avoid advanced indexing
                        append_dims = True
                        is_filter = True
                        if len(new_indexes[-1][1]) > 0:
                            new_indexes.append(([], []))
                    elif isinstance(index, slice):
                        append_dims = True

                if len(new_index_shape) > 0:
                    new_index = list(new_index)

                if state["need_dims"] and append_dims:
                    state["new_dims"].append(new_dim)
                if append_dims:
                    new_indexes[-1][0].append(slice(None))
                new_indexes[-1][1].append(new_index)
                if is_filter:
                    new_indexes.append(([], []))
                cur_dim += 1

        if new_indexes[-1][1] == []:
            new_indexes = new_indexes[:-1]

        return state["new_dims"], new_indexes

    def __getitem__(self, index):
        # try as a field name
        field_maybe = isinstance(index, six.string_types) and \
                      len(self.dtype) > 0
        if field_maybe and index in self.dtype.names:
            return numpy.ndarray.__getitem__(self, index)

        # try as an index
        try:
            dims, indexes = self._get_indexes(index, True)
        except IndexError as e:
            if field_maybe:
                # mimic numpy's field indexing error (ValueError)
                _, _, exc_traceback = sys.exc_info()
                new_e = ValueError(
                    ("field or tick named %s not found\n" % index) +
                    ("%s: %s" % (e.__class__.__name__, e)))
                six.reraise(new_e.__class__, new_e, exc_traceback)
            else:
                raise

        # nested indexing to support filters
        source = self
        prefix = []
        for idx_prefix, idx in indexes:
            idx = prefix + idx
            # trim tail empty slices
            while len(idx) > 0 and \
                  isinstance(idx[-1], slice) and \
                  idx[-1] == slice(None):
                idx = idx[:-1]
            if len(idx) > 0:
                if len(idx) == 1:
                    idx = idx[0]
                else:
                    idx = tuple(idx)
                res = source = numpy.ndarray.__getitem__(source, idx)
            else:
                res = source = numpy.ndarray.__getitem__(source, slice(None))
            prefix += idx_prefix

        if dims is None:
            res._dims = DataDims([])
            res._dims._dims = [meta.Dim(None, None)] * len(res.shape)
            res._dims._array = res

        elif len(res.shape) > 0:       # could be a record
            # avoid Dim copies
            # (no need to check variable consistency)
            res._dims = DataDims([])
            res._dims._dims = dims
            res._dims._array = res

            self_base = self.base if self.base is not None else self
            res_base = res.base if res.base is not None else res
            if res_base is not self_base:
                # copy dims: advanced indexing always returns a copy
                for idx, d in enumerate(res.dims):
                    if d.base is not None:
                        res._dims._dims[idx] = d.copy()
                        res._dims._dims[idx]._dims = res.dims
                    assert res.dims[idx].base is None

            for d in res.dims:
                d._dims = res._dims
            _ensure_shape(res)

        return res

    def __getslice__(self, start, stop):
        if start is 0:
            start = None
        if stop == sys.maxint:
            return self[start::]
        else:
            return self[start:stop:]

    def __setitem__(self, index, value):
        # set field
        if isinstance(index, six.string_types) and \
           len(self.dtype) > 0 and \
           index in self.dtype.names:
            numpy.ndarray.__setitem__(self.view(np.ndarray), index, value)
            return

        _, indexes = self._get_indexes(index, False)
        indexes = [i for _, idx in indexes for i in idx]
        numpy.ndarray.__setitem__(self.view(np.ndarray), tuple(indexes), value)

    def __setslice__(self, start, stop, value):
        if start is 0:
            start = None
        if stop == sys.maxint:
            self[start::] = value
        else:
            self[start:stop:] = value

    ##########################################################################
    # Miscellaneous methods
    ##########################################################################

    def _imag(self):
        res = self.view(numpy.ndarray).imag
        return data_array(res, dims=self)

    imag = property(_imag, doc=numpy.ndarray.imag.__doc__)

    def _real(self):
        res = self.view(numpy.ndarray).real
        return data_array(res, dims=self)

    real = property(_real, doc=numpy.ndarray.imag.__doc__)

    def ravel(self, order="C"):
        if order == "K":
            raise ValueError("Ravel with order 'K' not supported")

        res = self.view(numpy.ndarray).ravel(order)
        return data_array(res, dims=_flat_dimensions(self, order))

    ravel.__doc__ = numpy.ndarray.ravel.__doc__

    def transpose(self, *axes):
        res = numpy.ndarray.transpose(self, *axes)
        dims = self.dims
        if len(axes) == 0 or (len(axes) == 1 and axes[0] is None):
            dims = dims[::-1]
        else:
            if len(axes) == 1 and isinstance(axes[0], collections.Iterable):
                axes = axes[0]
            new_dims = list(dims)
            for i, ax in enumerate(axes):
                new_dims[i] = dims[ax]
            dims = new_dims
        return data_array(res, dims=dims)

    transpose.__doc__ = numpy.ndarray.transpose.__doc__

    def _T(self):
        if self.ndim < 2:
            return self
        return self.transpose()

    T = property(_T, doc=numpy.ndarray.T.__doc__)

    def flatten(self, order='C'):
        res = numpy.ndarray.flatten(self, order)
        return data_array(res, dims=_flat_dimensions(self, order))

    flatten.__doc__ = numpy.ndarray.flatten.__doc__

    ##########################################################################
    # Patched methods
    ##########################################################################

    mean = wrap_reduce(numpy.ndarray.mean)
    var = wrap_reduce(numpy.ndarray.var)
    std = wrap_reduce(numpy.ndarray.std)

    min = wrap_reduce(numpy.ndarray.min)
    max = wrap_reduce(numpy.ndarray.max)

    sum = wrap_reduce(numpy.ndarray.sum)
    prod = wrap_reduce(numpy.ndarray.prod)

    all = wrap_reduce(numpy.ndarray.all)
    any = wrap_reduce(numpy.ndarray.any)

    argmax = wrap_reduce(numpy.ndarray.argmax)
    argmin = wrap_reduce(numpy.ndarray.argmin)

    cumsum = wrap_accumulate(numpy.ndarray.cumsum)
    cumprod = wrap_accumulate(numpy.ndarray.cumprod)

    def __eq__(self, other):
        # TODO: handle metadata (check "Notes")
        return numpy.ndarray.__eq__(self.view(numpy.ndarray), other)

    def choose(self, choices, out=None, mode='raise'):
        res = numpy.ndarray.choose(self, choices, out=out, mode=mode)
        if out is not None:
            return res
        found = False
        for c in choices:
            if isinstance(c, Data):
                if c.dtype == res.dtype:
                    found = True
                break
        if not found:
            res = numpy.array(res)
        return res
    choose.func_doc = numpy.ndarray.choose.__doc__


def _patched_wrapper(name):
    """Wrap unhandled `numpy.ndarray` methods to return arrays without metadata.

    """
    func = getattr(numpy.ndarray, name)
    @functools.wraps(func, assigned=["__name__"])
    def wrapper(self, *args, **kwargs):
        ndself = self.view(numpy.ndarray)
        res = func(ndself, *args, **kwargs)
        return data_array(res)
    wrapper.__doc__ = "Wrapper for `numpy.ndarray.%s`." % name
    return wrapper


def _patch_methods():
    for name, obj in inspect.getmembers(Data):
        if inspect.getmodule(obj) == inspect.getmodule(Data):
            continue
        if name[0] == "_":
            continue
        if not callable(obj):
            continue
        # whitelist
        if name in [
                "byteswap", "dump", "fill", "getfield", "item", "itemset",
                "newbyteorder", "pformat", "put", "setfield", "setflags",
                "tolist", "tostring", "view"
        ]:
            continue
        setattr(Data, name, _patched_wrapper(name))

_patch_methods()


def _flat_dimensions(data, order):
    dims = data.dims

    if order == "A":
        order = "F" if numpy.isfortran(data) else "C"
    reverse = (order == "F")
    if reverse:
        dims = dims[::-1]

    expression = "-".join([dim.expression for dim in dims])

    def recurse(current, dims, result, reverse):
        if len(dims) == 0:
            if reverse:
                result.append("-".join(reversed(current)))
            else:
                result.append("-".join(current))
        else:
            if reverse:
                rest = dims[:-1]
                dim = dims[-1]
            else:
                rest = dims[1:]
                dim = dims[0]
            rest = dims[1:]
            dim = dims[0]
            for elem in dim:
                recurse(current + [elem], rest, contents, reverse)

    contents = []
    recurse([], dims, contents, reverse)

    return [meta.Dim(expression, contents)]


######################################################################
# * DataDims

class DataDims (pp.Pretty):
    """A collection of data dimensions that can be operated like a list.

    Parameters
    ----------
    values : DataDims, Data or list(meta.Dim)

    Notes
    -----
    The `values` argument can be a `DataDims` or `Data` object, in which case
    their metadata will be used.

    If `values` is a sequence, each element can be:

    * a `meta.Dim` object
    * a tuple with arguments for the `meta.Dim` constructor
    * a string with the first argument for the `meta.Dim` constructor

    Examples
    --------
    Dimensions are represented as a list:

    >>> dims = DataDims([('@a@-@b@', ['0-0']), ('@c@', ['0'])])
    >>> dims
    [Dim('@a@-@b@', ['0-0']), Dim('@c@', ['0'])]

    Dimensions can be retrieved using numeric indexes:

    >>> dims[0]
    Dim('@a@-@b@', ['0-0'])
    >>> dims[1]
    Dim('@c@', ['0'])

    Strings can also be used, referring to a variable name in a dimension, or
    the expression of such dimension:

    >>> dims['a']
    Dim('@a@-@b@', ['0-0'])
    >>> dims['b']
    Dim('@a@-@b@', ['0-0'])
    >>> dims['@a@-@b@']
    Dim('@a@-@b@', ['0-0'])
    >>> dims['c']
    Dim('@c@', ['0'])
    >>> dims['@c@']
    Dim('@c@', ['0'])

    Multiple dimensions can also be selected:

    >>> dims['b':]
    [Dim('@a@-@b@', ['0-0']), Dim('@c@', ['0'])]
    >>> dims['c', 'a']
    [Dim('@c@', ['0']), Dim('@a@-@b@', ['0-0'])]

    And dimensions can also be re-assigned:

    >>> dims['c'] = meta.Dim('@z@', ['foo'])
    >>> dims
    [Dim('@a@-@b@', ['0-0']), Dim('@z@', ['foo'])]

    As well as deleted:

    >>> del dims['z']
    >>> dims
    [Dim('@a@-@b@', ['0-0'])]

    """

    def __init__(self, values):
        self.__array = None
        if isinstance(values, DataDims):
            self._dims = np.ndarray(len(values), dtype=object)
            self._dims[:] = values.copy()
        elif isinstance(values, Data):
            self._dims = np.array(list(values.dims.copy()), dtype=object)
        else:
            self._dims = np.array([None] * len(values), dtype=object)
            self._dims[:] = [[]] * len(values)
            self[:] = values

    @property
    def _array(self):
        if self.__array is None:
            return None
        else:
            return self.__array()

    @_array.setter
    def _array(self, array):
        self.__array = weakref.ref(array)

    def __eq__(self, other):
        return isinstance(other, DataDims) and \
            len(self) == len(other) and \
            all(d1._expression == d2._expression and \
                len(d1) == len(d2) and \
                list(d1.ticks) == list(d2.ticks)
                for d1, d2 in zip(self, other))

    def _repr_pretty_(self, p, cycle):
        return p.pretty(list(self._dims))

    def __repr__(self):
        return pp.Pretty.__repr__(self)

    def copy(self):
        """Return a list with a deep copy of the dimension metadata."""
        return DataDims([d.copy() for d in self._dims])

    def __iter__(self):
        return iter(self._dims)

    def __len__(self):
        return len(self._dims)

    def _get_index_one(self, index, is_slice=False):
        if index is None and is_slice:
            return None
        elif isinstance(index, int):
            return index
        elif isinstance(index, tuple):
            return [self._get_index_one(idx)
                    for idx in index]
        elif isinstance(index, six.string_types):
            for idx, dim in enumerate(self._dims):
                if not dim._have_expression:
                    continue
                if index in dim.variables:
                    return idx
                if index == dim.expression:
                    return idx
            raise ValueError(
                "no dimension with given expression or variable name: %r" %
                index)
        else:
            raise TypeError("invalid dimension index: %r" % index)

    def _get_index(self, index):
        if isinstance(index, slice):
            return slice(
                self._get_index_one(index.start, is_slice=True),
                self._get_index_one(index.stop, is_slice=True),
                index.step)
        elif isinstance(index, list):
            return [self._get_index_one(idx)
                    for idx in index]
        else:
            return self._get_index_one(index)

    def __getitem__(self, index):
        idx = self._get_index(index)
        res = self._dims[idx]
        if isinstance(res, np.ndarray):
            return list(res)
        else:
            return res

    def __setitem__(self, index, value):
        idx = self._get_index(index)
        if isinstance(value, Data):
            value = value.dims
        if isinstance(value, DataDims):
            value = list(value)

        def as_dim(value, old_size):
            if value is None:
                res = meta.Dim(None, [None] * old_size)
            elif isinstance(value, meta.Dim):
                def array_base(array):
                    if array is not None and array.base is not None:
                        return array.base
                    else:
                        return array
                # use a view for all Dim objects of a view to this array
                if value.array is self._array:
                    res = value
                elif array_base(value.array) is array_base(self._array):
                    res = value._view(slice(None))[0]
                else:
                    res = value.copy()
            elif isinstance(value, six.string_types):
                res = meta.Dim(value, [None] * old_size)
            elif isinstance(value, collections.Sized) and len(value) == 2:
                contents = value[1]
                if contents is None:
                    contents = [None] * old_size
                res = meta.Dim(value[0], contents)
            else:
                raise TypeError("cannot build a dimension from: %s" %
                                repr(value))
            res._dims = self
            return res

        # apply changes

        new_dims = np.ndarray((len(self._dims),), dtype=object)
        new_dims[:] = self._dims

        if self._array is None:
            tmp_dims = new_dims[idx]
            if isinstance(tmp_dims, meta.Dim):
                old_sizes = [len(tmp_dims)]
            else:
                old_sizes = [len(o) for o in tmp_dims]
        else:
            old_sizes = np.array(self._array.shape)[idx]

        if isinstance(idx, int):
            idx = slice(idx, idx+1)
            value = [as_dim(value, old_sizes)]
        elif isinstance(idx, list) or isinstance(idx, slice):
            try:
                value_len = len(value)
            except TypeError:
                value_len = None
            if value_len != len(old_sizes):
                raise ValueError(
                    "invalid number of dimension metadata objects")
            value = [as_dim(v, o) for v, o in zip(value, old_sizes)]
        else:
            raise NotImplementedError(idx)

        # check we're not replacing views
        old_values = list(new_dims[idx])
        for dim in old_values:
            if dim == []:
                continue
            if dim.base is not None:
                raise ValueError("cannot replace a view")

        if isinstance(idx, list):
            # XXX: assigning to a numpy.ndarray results in None's
            for i, v in zip(idx, value):
                new_dims[i] = v
        else:
            new_dims[idx] = value

        # verify metadata consistency
        if self._array is not None:
            if len(new_dims) != self._array.ndim:
                raise ValueError(
                    "invalid number of dimension metadata objects")
        for i, d in enumerate(new_dims):
            if self._array is not None and \
               not (d == []) and\
               len(d) != self._array.shape[i]:
                raise ValueError("metadata for dimension %d with length %d "
                                 "does not match array dimension length %d" %
                                 (i, len(d), self._array.shape[i]))
            try:
                variables = d.variables
            except (meta.DimExpressionError, AttributeError):
                continue
            for var in variables:
                for pi, pd in enumerate(new_dims[:i]):
                    try:
                        pd_variables = pd.variables
                    except meta.DimExpressionError:
                        continue
                    if var in pd_variables:
                        msg = "expression variable %r " % var
                        msg += "in dimension %d " % i
                        msg += "already present in dimension %d" % pi
                        raise ValueError(msg)

        # update views
        for old_dim, new_dim in zip(old_values, value):
            if not isinstance(old_dim, meta.Dim):
                for idx, d in enumerate(self._dims):
                    if d == old_dim:
                        break
                self._dims[idx] = new_dim
            elif old_dim is not new_dim:
                old_dim._assign(new_dim)

    def __getslice__(self, start, stop):
        if start is 0:
            start = None
        if stop == sys.maxint:
            return self[start::]
        else:
            return self[start:stop:]

    def __setslice__(self, start, stop, value):
        if start is 0:
            start = None
        if stop == sys.maxint:
            self[start::] = value
        else:
            self[start:stop:] = value

    def __delitem__(self, index):
        if self._array is not None:
            raise TypeError(
                "cannot delete a dimension from an existing Data array")
        idx = self._get_index(index)
        if not isinstance(idx, collections.Iterable):
            idx = [idx]
        new_dims = [dim
                    for i, dim in enumerate(self._dims)
                    if i not in idx]
        self._dims = new_dims

    def __getstate__(self):
        return {"_dims": self._dims}

    def __setstate__(self, odict):
        self.__dict__.update(odict)
        self.__array = None
        for d in self._dims:
            d._dims = self


######################################################################
# * DataIndexer

class DataIndexer(pp.Pretty):
    """Object to build an array indexing tuple across dimensions.

    This comes in handy when indexing an array without caring about the precise
    position of every dimension, and can be used to index `Data` objects.

    The `dims` attribute returns a "dimension selector", which can be indexed
    to select dimensions. Indexing the result will generate a new `DataIndexer`
    referencing the selected dimensions.


    Examples
    --------
    >>> d = data_array(numpy.arange(27).reshape((3, 3, 3)),
    ...                dims=[("@%s@" % d, ["%s%d" % (d, i) for i in range(3)])
    ...                      for d in ["a", "b", "c"]])

    Regular `Data` indexing would be:

    >>> d["a0", :, "c=='c1'":]
    data_array([[1, 2], [4, 5], [7, 8]],
               dims=[('@b@', ['b0', 'b1', 'b2']), ('@c@', ['c1', 'c2'])])

    Using `DataIndexer` allows ignoring dimension positions (useful to
    programmatically build indexing objects):

    >>> i = d.indexer
    >>> i = i.dims["a"]["a0"]
    >>> i = i.dims["c"]["c == 'c1'":]
    >>> i
    ('a0', slice(None, None, None), slice("c == 'c1'", None, None))
    >>> d[i]
    data_array([[1, 2], [4, 5], [7, 8]],
               dims=[('@b@', ['b0', 'b1', 'b2']), ('@c@', ['c1', 'c2'])])

    Multiple dimensions can also be used, resulting in the expected results:

    >>> i.dims["b"][0, np.newaxis]
    ('a0', 0, None, slice("c == 'c1'", None, None))
    >>> i.dims["c", "a"][-1, 1]
    (1, slice(None, None, None), -1)
    >>> i.dims["c", "c"][-1, 1]
    ('a0', slice(None, None, None), 1)

    You can also obtain indexes compatible with regular numpy arrays:


    >>> d.indexer.dims["a", "b", "c"]["a0", 0, np.newaxis, "c == 'c1'":].as_tuple()
    (0, 0, None, slice(1, None, None))


    Notes
    -----
    Instances of this class should be obtained through `Data.indexer`.


    See also
    --------
    Data.indexer
    Data.idata

    """

    def __init__(self, array=None, dims=None, indexes=None):
        assert ((array is not None and dims is None and indexes is None) or
                (array is None and dims is not None and indexes is not None))
        if dims is None:
            # NOTE: need the complete dimension contents for 'as_tuple'
            self._dims = array.dims
        else:
            self._dims = dims
        if indexes is None:
            self._indexes = [slice(None) for index in range(array.ndim)]
        else:
            self._indexes = list(indexes)

    def _get_indexes(self, trim=True):
        if trim:
            indexes = _indexes_trim_suffix(self._indexes)
        else:
            indexes = self._indexes
        if len(indexes) == 1:
            return indexes[0]
        else:
            return tuple(indexes)

    def as_tuple(self, trim=True):
        """Return a tuple with regular numpy indexes.

        Parameters
        ----------
        trim : bool, optional
            Omit indexing information for trailing dimensions not referenced on
            the indexer object (default: `True`).

        """
        idxs = self._get_indexes(trim=trim)
        if not isinstance(idxs, tuple):
            idxs = tuple([idxs])
        assert len(idxs) <= len(self._dims)
        res = []
        for dim, idx in zip(self._dims, idxs):
            if idx is np.newaxis:
                res.append(idx)
            else:
                res.append(dim._get_indexes(idx))
        if len(res) == 1:
            return res[0]
        else:
            return tuple(res)

    def __getitem__(self, index):
        return self.dims[:][index]

    @property
    def dims(self):
        """Proxy to indexing the selected dimensions (see `DataDims`)."""
        return _DataIndexer_Dims(self)

    def _repr_pretty_(self, p, cycle):
        return p.pretty(self._get_indexes(trim=False))

    def __repr__(self):
        return pp.Pretty.__repr__(self)


class _DataIndexer_Dims(object):
    def __init__(self, indexer):
        self._indexer = indexer

    def __getitem__(self, index):
        dims = self._indexer._dims[index]
        if isinstance(dims, meta.Dim):
            dims = [dims.index]
        else:
            dims = [dim.index for dim in dims]
        return _DataIndexer_DimsIndexer(self._indexer, dims)


class _DataIndexer_DimsIndexer(object):
    def __init__(self, indexer, dims):
        self._indexer = indexer
        self._dims = dims

    def __getitem__(self, index):
        indexes = _indexes_normalize(index, len(self._dims))

        i_dims = list(self._indexer._dims.copy())
        i_indexes = list(self._indexer._indexes)

        # set indexing info
        pos = 0
        for idx in indexes:
            if idx is np.newaxis:
                continue
            i_indexes[self._dims[pos]] = idx
            pos += 1

        # extend with newaxis
        pos -= 1
        for idx in reversed(indexes):
            if idx is not np.newaxis:
                pos -= 1
            else:
                if pos == -1:
                    dims_pos = self._dims[pos+1]
                else:
                    dims_pos = self._dims[pos] + 1
                i_dims = (i_dims[:dims_pos] +
                          [meta.Dim()] +
                          i_dims[dims_pos:])
                i_indexes = (i_indexes[:dims_pos] +
                             [np.newaxis] +
                             i_indexes[dims_pos:])

        return DataIndexer(dims=DataDims(i_dims),
                           indexes=i_indexes)


######################################################################
# * DataIndexerProxy

class DataIData(object):
    """Object to index specific array dimensions without knowing their position.

    This comes in handy as a shorthand for `DataIndexer` when the indexer
    object does not need to be reused.


    Examples
    --------
    >>> d = data_array(numpy.arange(27).reshape((3, 3, 3)),
    ...                dims=[("@%s@" % d, ["%s%d" % (d, i) for i in range(3)])
    ...                      for d in ["a", "b", "c"]])

    Shorthand of the equivalent first example in `DataIndexer`:

    >>> d.idata["a", "c"]["a0", "c == 'c1'":]
    data_array([[1, 2], [4, 5], [7, 8]],
               dims=[('@b@', ['b0', 'b1', 'b2']), ('@c@', ['c1', 'c2'])])

    You can also chain it multiple times:

    >>> d.idata["a"]["a0"].idata["c"]["c == 'c1'":]
    data_array([[1, 2], [4, 5], [7, 8]],
               dims=[('@b@', ['b0', 'b1', 'b2']), ('@c@', ['c1', 'c2'])])


    Notes
    -----
    Instances of this class should be obtained through `Data.idata`.

    Chaining this functionality multiple times (second example) creates a
    temporary result array on each step. This must be kept in mind when using
    it to assign new values, since "advanced indexing" (see NumPy documentation
    [#np.idx]_) creates copies of the data instead of a view (i.e., the
    assigned values will not reach the original array).

    .. [#np.idx] http://docs.scipy.org/doc/numpy/reference/arrays.indexing.html


    See also
    --------
    Data.idata

    """

    def __init__(self, array, indexes=None):
        self._array = array

    def __getitem__(self, index):
        indexer = self._array.indexer.dims[index]
        return _DataIData_DataIndexer(self._array, indexer)


class _DataIData_DataIndexer(object):

    def __init__(self, array, indexer):
        self._array = array
        self._indexer = indexer

    def __getitem__(self, index):
        indexer = self._indexer[index]
        return self._array[indexer]

    def __setitem__(self, index, value):
        indexer = self._indexer[index]
        self._array[indexer] = value


######################################################################
# * Public manifest

from ._funcs import *


__all__ = [
    "data_array", "data_frombuffer", "data_memmap",
    "Data", "DataDims", "DataIndexer", "DataIData",
    "wrap_reduce", "wrap_accumulate",
]

__all__ += _funcs.__all__



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Dimension metadata containers."""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2009-2015, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import collections
import numbers
import numpy
import re
import six
import sys
import weakref

from sciexp2.common import progress
import sciexp2.common.filter
from sciexp2.common import pp
from sciexp2.common import varref
import sciexp2.common.instance
from sciexp2.common.instance import Instance


class DimExpressionError (Exception):
    """Invalid dimension expression."""
    def __init__(self, *args, **kwargs):
        super(DimExpressionError, self).__init__(*args, **kwargs)


class DimExpression(object):
    """Descriptor for a dimension expression

    Parameters
    ----------
    expression : str
        Expression describing each element.
    varre : str, optional
        Regular expression for variables in `expression` (default in "@"; see
        `sciexp2.common.varref.build_varre`).
    validate : list of str, optional
        Ensure all variables in `expression` appear in `validate`.

    Raises
    ------
    DimExpressionError
        Invaid expression.

    Examples
    --------
    >>> e = DimExpression("@a@ (@b@)")
    >>> e.variables
    ['a', 'b']
    >>> e.expression
    '@a@ (@b@)'
    >>> e.extract('1 (2)')
    Instance({'a': 1, 'b': 2})
    >>> e.translate({'a': 1, 'b': 2})
    '1 (2)'

    """
    def __init__(self, expression, varre=None, validate=None):
        if isinstance(expression, DimExpression):
            self.expression = str(expression.expression)
            if varre is not None:
                msg = "Cannot provide a DimExpression and varre at the same time"
                raise DimExpressionError(msg)
            self.varre = expression.varre
        else:
            self.expression = expression
            if varre is None:
                varre = "@"
            self.varre = varre

        def escape(expression):
            expr = expression
            expr = re.escape(expr).replace("\\@", "@")
            tokenizer = varref.Tokenizer(self.varre)
            res = []
            for is_variable, token in tokenizer.tokenize(expr):
                if is_variable:
                    # unescape non-alphanumeric characters in variable names
                    res.append("@%s@" % token.replace("\\", ""))
                else:
                    res.append(token)
            expr = "".join(res)
            if expr[-1] != "$":
                expr += "$"
            return expr
        true_expression = escape(self.expression)

        self.variables = varref.expr_get_vars(self.varre, self.expression)

        expr_regex = varref.expr_to_regexp(self.varre, self.expression)
        regex_vars = varref.regexp_get_names(expr_regex)
        if len(set(regex_vars) - set(self.variables)) > 0:
            msg = "A dimension expression cannot contain regexp named groups: %s"
            raise DimExpressionError(msg % self.expression)

        if len(self.variables) == 0:
            raise DimExpressionError(
                "Dimension expression contains no variables: " +
                str(self.expression))

        if validate is not None:
            for var in self.variables:
                if var not in validate:
                    msg = "Unknown variable %r in dimension expression: %s"
                    raise DimExpressionError(msg % (var, self.expression))

        self._xlator = varref.PrepDictXlator(self.varre)
        self._xlator.prepare(self.expression)

        self._extractor = sciexp2.common.instance.InstanceExtractor(
            true_expression, self.varre)

    def __eq__(self, other):
        return isinstance(other, DimExpression) and \
            self.expression == other.expression and \
            self.varre == other.varre

    def extract(self, source):
        """Extract an instance from the given string."""
        instance = self._extractor.extract(source)
        if instance is None:
            raise ValueError("Cannot extract expression %r from %r" %
                             (self.expression, source))
        # try to turn values into numbers
        for k, v in instance:
            instance[k] = sciexp2.common.utils.str2num(v)
        return instance

    def translate(self, instance):
        """Translate the given instance into a string."""
        return self._xlator.xlate(None, instance)


class Dim (pp.Pretty):
    """Map variable values to and from array indexes.

    Parameters
    ----------
    expression : str, optional
        Expression describing each element.
    contents : optional, optional
        Dimension contents (default is no contents).

    Attributes
    ----------
    expression
    variables
    indexes
    ticks
    instances
    values
    unique_values
    array
    index

    Notes
    -----
    If `expression` is not given, it will default to ``@var@``.

    Each element of `contents` must have one of the following formats:

    * `None`: the entry is left empty.

    * `str`: an instance will be extracted using the `expression` attribute.

    * `dict`: a dict-like object that must have, at least, entries for the
      variables in the `expression` attribute.

    When not `None`, any pair of elements must have a different expansion for
    `expression`.

    A dimension can be indexed with one of the following formats:

    * `int`: element at given position.

    * `str`: element with given tick.

    * `dict`: element with given variable values (can also be an `Instance`).

    If the stepping of a slice is a `str` or `dict`, it will be used as a
    filter to select the matching elements. The `dict` version is a performance
    optimization for filters that look for the specified variable values.

    Accessing a single element of a `Dim` returns None, but slicing it results
    in either a view to the original, or a brand new object. The type of result
    depends on the indexing used to slice it, following numpy's slicing rules.
    Also, using a filter while slicing always creates a new object.

    Assigning to a `Dim` object follows the same rules, and accepts the same
    values valid during construction.

    Examples
    --------

    >>> d = Dim('@a@-@b@', ['a1-b1', 'a1-b2', None, 'a2-b2', None])
    >>> d[1:-1]
    Dim('@a@-@b@', ['a1-b2', None, 'a2-b2'])
    >>> d['a1-b2':-1]
    Dim('@a@-@b@', ['a1-b2', None, 'a2-b2'])

    The stepping in a slice is used as a filter:

    >>> d[::"a == 'a1'"]
    Dim('@a@-@b@', ['a1-b1', 'a1-b2'])

    Using a `dict` (or `Instance`) is a performance-optimized version of
    indexing and filtering:

    >>> d[{'a': 'a1', 'b':'b2'}:-1]
    Dim('@a@-@b@', ['a1-b2', None, 'a2-b2'])
    >>> d[::{'a': 'a1'}]
    Dim('@a@-@b@', ['a1-b1', 'a1-b2'])

    Indexing on a single element returns `None`:

    >>> d[0] is d['a1-b1'] is d[1] is None
    True

    The same formats (`str` or `dict`) are supported when testing for
    existence:

    >>> 'a1-b1' in d
    True
    >>> Instance({'a': 'a1', 'b': 'b1'}) in d
    True
    >>> 'a1-b3' in d
    False
    >>> Instance({'a': 'a1', 'b': 'b3'}) in d
    False

    See also
    --------
    base

    """

    ##########################################################################
    # Construction
    ##########################################################################

    _DEFAULT_VAR = "var"
    _DEFAULT_EXPR = "@%s@" % _DEFAULT_VAR

    def __init__(self, expression=None, contents=None):
        # weakref to containing DataDims
        self.__dims = None
        # ultimate base Dim for which this is a view for (possibly indirectly)
        self._base = None
        # views of self (id(Dim), weakref(Dim)) (both direct and indirect)
        self._views = {}

        if expression is None:
            self._have_expression = False
            expression = self._DEFAULT_EXPR
        else:
            self._have_expression = True
        self._expression = DimExpression(expression)

        if contents is None:
            contents = []

        # fill-in contents
        self._group = sciexp2.common.instance.InstanceGroup(
            view_able=True,
            cache=True,
            cache_proactive=[],
            cache_others=self._expression.variables)
        self._indexes = {}

        if not isinstance(contents, progress.Progressable):
            contents = progress.progressable_simple(
                contents, None,
                msg="building dimension...")
        expressions = dict()
        with contents:
            for i in contents:
                instance, i_expression = self._as_instance(i)

                # check for repeated instances
                # NOTE: cannot use '_exists_instance' since we want to avoid
                #       using caching
                if len(instance) > 0:
                    if i_expression is None:
                        i_expression = self._expression.translate(instance)
                    if i_expression in expressions:
                        raise ValueError(
                            "repeated tick %r while adding %r" %
                            (i_expression, instance))
                    expressions[i_expression] = instance

                self._group.add(instance)
                self._indexes[instance] = len(self._indexes)

    def _exists_instance(self, instance, ignore=set()):
        for i in self._group.select(instance):
            if i not in ignore:
                return True
        return False

    def _as_instance(self, content):
        if isinstance(content, six.string_types):
            i_expression = content
            content = self._expression.extract(content)
        else:
            i_expression = None

        instance = Instance()
        if content is not None:
            try:
                instance = sciexp2.common.instance.Instance(
                    (key, val) for key, val in six.iteritems(content)
                    if key in self._expression.variables)
            except AttributeError:
                raise TypeError("dont know how to construct "
                                "a dimension element from: %r" % content)

        # check for missing variables
        missing = set(self._expression.variables) - set(instance.keys())
        if len(missing) > 0 and len(instance) > 0:
            raise KeyError("missing variable %r for: %r" %
                           (missing.pop(), content))

        return instance, i_expression

    @property
    def _dims(self):
        if self.__dims is None:
            return None
        else:
            return self.__dims()

    @_dims.setter
    def _dims(self, value):
        if value is None:
            self.__dims = None
        else:
            self.__dims = weakref.ref(value)

    def copy(self):
        """Return a deep copy of this dimension."""
        return Dim(self.expression if self._have_expression else None,
                   self._group)

    @property
    def expression(self):
        """An expandable expression where dimension elements are built from.

        Raises
        ------
        DimExpressionError
            Non-existing dimension expression.

        See also
        --------
        DimExpression

        Notes
        -----
        If a new value is assigned, ticks will be kept consistent.

        """
        if self._have_expression:
            return str(self._expression.expression)
        else:
            raise DimExpressionError("dimension has no expression")

    @expression.setter
    def expression(self, expression):
        have_expression = expression is not None
        if not have_expression:
            expression = self._DEFAULT_EXPR
        expression = DimExpression(expression)

        def change_expression(have_expression, expression):
            base = self.base or self
            base._have_expression = have_expression
            base._expression = expression
            for view in base._views.values():
                view()._have_expression = have_expression
                view()._expression = expression

        if self._have_expression:
            if have_expression:
                if set(expression.variables) != set(self.variables):
                    raise DimExpressionError(
                        "variable mismatch "
                        "(you can set 'expression' to None to reset)")
                change_expression(True, expression)
            else:
                base = self.base or self
                ticks = list(base.ticks)
                change_expression(False, expression)
                for tick, instance in zip(ticks, base._group):
                    instance.clear()
                    if tick is not None:
                        instance[self._DEFAULT_VAR] = tick
                self._group.cache_reset(hard=True)
                self._group.cache_set(True, [], expression.variables)
        else:
            base = self.base or self
            base_expression = base._expression
            change_expression(have_expression, expression)

            # calculate new instances; revert on error
            instances = []
            for instance in self._group:
                if len(instance) > 0:
                    try:
                        instance = self._as_instance(str(instance[self._DEFAULT_VAR]))[0]
                    except:
                        change_expression(False, base_expression)
                        raise
                instances.append(instance)

            # apply new instances
            for cur_instance, new_instance in zip(base._group, instances):
                cur_instance.clear()
                cur_instance.update(new_instance)
            self._group.cache_reset(hard=True)
            self._group.cache_set(True, [], expression.variables)

    def extract(self, source):
        """Return an instance from `source` using the dimension's `expression`.

        Examples
        --------
        >>> d = Dim("@entries@-@entry_size@",
        ...         ["1-8", "2-8", "2-16"])
        >>> d.extract("10-20")
        Instance({'entries': 10, 'entry_size': 20})

        Raises
        ------
        DimExpressionError
            Non-existing dimension expression.

        """
        if self._have_expression:
            return self._expression.extract(source)
        else:
            raise DimExpressionError("dimension has no expression")

    def translate(self, instance):
        """Translate `instance` into a string using the dimension's `expression`.

        Examples
        --------
        >>> d = Dim("@entries@-@entry_size@",
        ...         ["1-8", "2-8", "2-16"])
        >>> d.translate(Instance({'entries': 10, 'entry_size': 20}))
        '10-20'

        Raises
        ------
        DimExpressionError
            Non-existing dimension expression.

        """
        if self._have_expression:
            return self._expression.translate(instance)
        else:
            raise DimExpressionError("dimension has no expression")

    @property
    def variables(self):
        """List of variable names in `expression`.

        Raises
        ------
        DimExpressionError
            Non-existing dimension expression.

        """
        if self._have_expression:
            return list(self._expression.variables)
        else:
            raise DimExpressionError("dimension has no expression")

    ##########################################################################
    # Translation API
    ##########################################################################

    def _get_indexes(self, index):
        """Return the numeric index translation of names in `index`."""

        ##################################################
        # numeric indexes
        if isinstance(index, numbers.Real):
            if index >= len(self._group):
                raise IndexError("index (%d) out of range" % index)
            return index

        # NOTE: PFilter confuses `numpy.shape`
        if isinstance(index, sciexp2.common.filter.PFilter):
            index = index.realize()

        ##################################################
        # sequence of indexes (need further parsing)
        if isinstance(index, numpy.ndarray) or numpy.shape(index) != tuple():
            if not isinstance(index, numpy.ndarray):
                tolist = True
                index = numpy.array(index, dtype=object)
            else:
                tolist = False

            if numpy.issubdtype(index.dtype, numpy.bool) and \
               numpy.issubdtype(index.dtype, numpy.bool_):
                if index.ndim == 0:
                    raise ValueError(
                        "boolean index array should have 1 dimension")
                return [self._get_indexes(pos)
                        for pos, idx in enumerate(index)
                        if idx]

            elif index.ndim == 0:
                return self._get_indexes(index)

            else:
                res = numpy.ndarray(index.shape, dtype=int)
                for pos, idx in enumerate(index.flat):
                    try:
                        elem = self._get_indexes(idx)
                    except Exception as e:
                        _, _, exc_traceback = sys.exc_info()
                        new_e = IndexError(
                            "invalid index element %r in position %d"
                            "\n  [was] "
                            "%s: %s" % (idx, pos,
                                        e.__class__.__name__, e))
                        six.reraise(new_e.__class__, new_e, exc_traceback)
                    res.flat[pos] = elem
                if tolist:
                    # NOTE: work around numpy bug #6564
                    #       https://github.com/numpy/numpy/issues/6564
                    res = res.tolist()
                return res

        ##################################################
        # slice
        if isinstance(index, slice):
            def slice_index(index):
                if index is None:
                    return index
                if isinstance(index, numbers.Real):
                    # integer slicing accepts out-of-bounds indexes
                    return index
                return self._get_indexes(index)
            start = slice_index(index.start)
            stop = slice_index(index.stop)
            step = index.step

            if not (start is None or isinstance(start, int)) or \
               not (stop is None or isinstance(stop, int)):
                raise IndexError("invalid slice (%s)" % repr(index))

            if step is None or isinstance(step, int):
                return slice(start, stop, step)

            # string stepping -> filter
            try:
                filter_ = sciexp2.common.filter.Filter(step)
                filter_.validate(self._group)
            except Exception as e:
                _, _, exc_traceback = sys.exc_info()
                new_e = IndexError(
                    "invalid slice filter\n"
                    "%s: %s" % (e.__class__.__name__, e))
                six.reraise(new_e.__class__, new_e, exc_traceback)

            if start is None:
                start = 0
            assert isinstance(start, numbers.Real)
            if stop is None:
                stop = len(self._group)
            elif stop < 0:
                stop = len(self._group) + stop
            assert isinstance(stop, numbers.Real)
            return [self._indexes[idx] for idx in self._group.select(
                filter_, allow_unknown=True)
                    if start <= self._indexes[idx] and self._indexes[idx] < stop]

        ##################################################
        # label
        if isinstance(index, six.string_types):
            instance = self._get_from_expression(index)
            if instance is not None:
                return self._indexes[instance]

        ##################################################
        # instance in this dimension
        if isinstance(index, dict):
            if isinstance(index, Instance):
                # shortcut when indexing with existing instance
                indexes = self._indexes.get(index, None)
                if indexes is not None:
                    return indexes
            indexes = self._get_indexes(slice(None, None, index))
            if len(indexes) == 1:
                return indexes[0]

        ##################################################
        # filter (or else error)
        try:
            filter_ = sciexp2.common.filter.Filter(index)
        except TypeError:
            raise TypeError("invalid index type: " + repr(index))
        except sciexp2.common.filter.ParseError as e:
            raise IndexError("invalid filter: %r\n  "
                             "[was] %s: %s" % (index, e.__class__.__name__,
                                               str(e)))
        try:
            selected = list(self._group.select(filter_))
        except sciexp2.common.filter.FilterError as e:
            if e.args[0] == index:
                raise IndexError("unknown tick: %r" % index)
            else:
                raise IndexError("invalid filter: %r\n  "
                                 "[was] %s: %s" % (index, e.__class__.__name__,
                                                   str(e)))
        if len(selected) == 0:
            raise IndexError("filter has no matches: %r" % index)
        if len(selected) == 1:
            return self._indexes[selected[0]]
        else:
            raise IndexError("filter has multiple matches: %r" % index)

    def _get_from_expression(self, expression):
        try:
            instance = self._expression.extract(expression)
        except ValueError:
            return None
        # try using an instance as a filter
        try:
            results = list(self._group.select(instance))
        except sciexp2.common.filter.ParseError:
            pass
        else:
            if len(results) > 0:
                assert len(results) == 1
                return results[0]

        # fall back to reconstructing expression
        for instance in self._group:
            try:
                i_expression = self._expression.translate(instance)
            except varref.XlateError:
                continue
            if expression == i_expression:
                return instance
        return None

    ##########################################################################
    # Slicing
    ##########################################################################

    def __getslice__(self, start, stop):
        if start is 0:
            start = None
        if stop == sys.maxint:
            return self[start::]
        else:
            return self[start:stop:]

    def _view(self, index):
        indexes = self._get_indexes(index)
        if isinstance(indexes, slice):
            dim = Dim(self.expression if self._have_expression else None,
                      [])
            dim._base = self._base or self
            self._views[id(dim)] = weakref.ref(dim)

            if indexes.start is None and indexes.stop is None and \
               indexes.step is None:
                dim._group = self._group
                dim._indexes = self._indexes
            else:
                dim._group = self._group.view(indexes)
                dim._indexes = dict(
                    six.moves.zip(dim._group,
                                  six.moves.range(len(dim._group))))
        elif numpy.shape(indexes) == tuple():
            dim = None
        else:
            indexes_ = indexes
            if not isinstance(indexes_, numpy.ndarray):
                indexes_ = numpy.array(indexes, dtype=int)
            if indexes_.ndim > 1:
                raise TypeError(
                    "dimension indexes cannot have more than one dimension")
            with progress.progressable_simple(
                    (self._group.get_index(idx)
                     for idx in indexes_.flat),
                    indexes_, msg="slicing dimension...") as contents:
                dim = Dim(self.expression if self._have_expression else None,
                          contents)
        return dim, indexes

    def _assign(self, new_dim):
        base = self.base or self

        # reassign contents
        assert len(self._group) == len(new_dim._group)
        for i_old, i_new in zip(base._group, new_dim._group):
            for key in list(i_old.keys()):
                del i_old[key]
            i_old.update(i_new)

        # update expressions
        base._have_expression = new_dim._have_expression
        base._expression = new_dim._expression
        base._group.cache_reset()
        base._group.cache_set(True, proactive=[],
                              others=base._expression.variables)
        for dim in self._views.values():
            dim()._have_expression = new_dim._have_expression
            dim()._expression = new_dim._expression
            dim()._group.cache_reset()
            dim()._group.cache_set(True, proactive=[],
                                   others=self._expression.variables)

    def __getitem__(self, index):
        return self._view(index)[0]

    def __setslice__(self, start, stop, value):
        if start is 0:
            start = None
        if stop == sys.maxint:
            self[start::] = value
        else:
            self[start:stop:] = value

    def __setitem__(self, index, value):
        indexes = self._get_indexes(index)

        # homogeinize 'value' and record 'targets'
        if isinstance(indexes, slice):
            start = indexes.start
            if start is None:
                start = 0
            elif start < 0:
                start += len(self._group)

            stop = indexes.stop
            if stop is None:
                stop = len(self._group)
            elif stop < 0:
                stop += len(self._group)

            step = indexes.step
            if step is None:
                step = 1

            targets = []
            idx = start
            while idx < stop:
                targets.append(self._group.get_index(idx))
                idx += step
        elif numpy.shape(indexes) == tuple():
            value = [value]
            targets = [self._group.get_index(indexes)]
        else:
            targets = [self._group.get_index(idx)
                       for idx in indexes]
        try:
            len_value = len(value)
        except TypeError:
            len_value = 0
        if len(targets) != len_value:
            raise ValueError(
                "cannot assign sequence of size %d to slice of size %d" %
                (len_value, len(targets)))

        # actual updates
        base = self.base or self
        base._group.cache_prefetch()
        ignore = set(targets)
        expressions = set()
        for target, i in zip(targets, value):
            new_instance, new_expression = self._as_instance(i)
            if len(new_instance) > 0:
                if new_expression is None:
                    new_expression = self._expression.translate(new_instance)
                if base._exists_instance(new_instance, ignore) or \
                   new_expression in expressions:
                    raise ValueError("repeated tick %r while adding %r" %
                                     (new_expression, new_instance))
                target.update(new_instance)
                expressions.add(new_expression)
            else:
                target.clear()

        # reset caches
        base._group.cache_reset()
        for view in base._views.values():
            base._group.cache_reset()

    def __del__(self):
        # NOTE: __del__ can be called more than once
        #       (if object is resurrected during __del__)
        if self.base is not None and id(self) in self._base._views:
            del self._base._views[id(self)]

    @property
    def base(self):
        """The `Dim` object this one is a view of."""
        return self._base

    @property
    def indexes(self):
        """Return a `DimViewIndexes` for this dimension."""
        return DimViewIndexes(self)

    @property
    def ticks(self):
        """Return a `DimViewTicks` for this dimension."""
        return DimViewTicks(self)

    @property
    def instances(self):
        """Return a `DimViewInstances` for this dimension."""
        return DimViewInstances(self)

    @property
    def values(self):
        """Return a `DimViewValues` for this dimension."""
        return DimViewValues(self, unique=False)

    @property
    def unique_values(self):
        """Return a `DimViewValues` for this dimension (without repeated values)."""
        return DimViewValues(self, unique=True)

    @property
    def array(self):
        """The `Data` array this dimension is associated to."""
        if self._dims is None:
            return None
        else:
            return self._dims._array

    @property
    def index(self):
        """The index of this dimension in the associated `DataDims`."""
        if self._dims is None:
            return None
        return list(self._dims).index(self)

    ##########################################################################
    # (un)pickling
    ##########################################################################

    def __getstate__(self):
        if self._have_expression:
            expression = self.expression
            varre = self._expression.varre
        else:
            expression = self._DEFAULT_EXPR
            varre = None
        odict = self.__dict__.copy()
        odict["_expression"] = expression
        odict["_varre"] = varre
        del odict["_%s__dims" % self.__class__.__name__]
        del odict["_base"]
        del odict["_views"]
        return odict

    def __setstate__(self, odict):
        self._expression = DimExpression(odict["_expression"], odict["_varre"])
        del odict["_expression"]
        del odict["_varre"]
        self.__dict__.update(odict)
        self._dims = None
        self._base = None
        self._views = {}

    ##########################################################################
    # Miscellaneous
    ##########################################################################

    def __contains__(self, index):
        """Return whether an elment named 'index' exists."""
        if isinstance(index, six.string_types) or isinstance(index, dict):
            try:
                self._get_indexes(index)
                return True
            except IndexError:
                return False
        else:
            raise TypeError("not a string or dict")

    def __len__(self):
        """The length of this dimension."""
        return len(self._group)

    def __iter__(self):
        for i in self._group:
            try:
                yield self._expression.translate(i)
            except varref.XlateError:
                yield None

    def _repr_pretty_(self, p, cycle):
        with self.pformat(p, cycle, name="Dim"):
            if self._have_expression:
                p.text(repr(self.expression))
            else:
                p.text(repr(None))
            p.text(",")
            p.breakable()
            p.pretty(list(self))

    def __repr__(self):
        return pp.Pretty.__repr__(self)

    def sorted(self, *args, **kwargs):
        """Same as `sort`, but returns a sorted copy."""
        res = self.copy()
        res.sort(*args, **kwargs)
        return res

    def sort(self, *spec):
        """Sort dimension elements in-place.

        Parameters
        ----------
        spec : list
            Sequence of sorting specifications.

        Raises
        ------
        DimExpressionError
            Dimension has no expression.

        Notes
        -----
        Each element in `spec` can have one of the following formats:

        * ``(str, callable)``
            Sort values of variable ``str`` using the `cmp`-like function
            ``callable``, accepting pairs of variable values.

        * ``(str, list)``
            Sort values of variable ``str`` according to the provided list.

        * ``(str, True)``
            Sort values of variable ``str`` lexicographically.

        * ``callable``
            Sort all dimension instances using the `cmp`-like function
            ``callable``, accepting pairs of dimension instances.

        * ``list``
            Sort all dimension instances accoring to the given tick list.

        * ``str``
            Sort values of variable ``str`` lexicographically.

        * ``True``
            Sort all dimension instances lexicographically.

        For every pair of instances, sorting will stop at the first `spec`
        element with a non-zero result.

        Sorting on a dimension view will only affect the elements contained in
        it (i.e., will affect views of that view, but not elements in the base
        dimension that the view is not referencing).

        Sorting on a dimension that has an associated array will also sort the
        array contents accordingly.

        """
        # default sorting functions
        def sort_instance_from_func(sort_func):
            def func(i1, i2):
                if len(i1) == 0:
                    i1 = None
                if len(i2) == 0:
                    i2 = None
                return sort_func(i1, i2)
            return func

        def sort_var_from_func(var, sort_func):
            def func(i1, i2):
                v1 = i1[var] if i1 is not None else None
                v2 = i2[var] if i2 is not None else None
                return sort_func(v1, v2)
            return sort_instance_from_func(func)

        def sort_var_from_list(contents, var):
            def func(v1, v2):
                idx1 = contents.index(v1)
                idx2 = contents.index(v2)
                return cmp(idx1, idx2)
            return sort_var_from_func(var, func)

        def sort_var_lex(var):
            return sort_var_from_func(var, cmp)

        def sort_tick_from_func(sort_func):
            def func(i1, i2):
                if i1 is None:
                    v1 = None
                else:
                    v1 = self.translate(i1)
                if 21 is None:
                    v2 = None
                else:
                    v2 = self.translate(i2)
                return sort_func(v1, v2)
            return sort_instance_from_func(func)

        def sort_tick_from_list(contents):
            def func(v1, v2):
                try:
                    idx1 = contents.index(v1)
                    idx2 = contents.index(v2)
                except:
                    raise
                return cmp(idx1, idx2)
            return sort_tick_from_func(func)

        # normalize 'spec'
        def get_variables(idx, elem):
            try:
                return self.variables
            except DimExpressionError as e:
                _, _, exc_traceback = sys.exc_info()
                new_e = DimExpressionError(
                    "spec %d references variables on empty expression: %r"
                    "\n  [was] "
                    "%s: %s" % (idx, elem,
                                e.__class__.__name__, e))
                six.reraise(new_e.__class__, new_e, exc_traceback)

        def normalize(spec):
            res = []
            for idx, s in enumerate(spec):
                elem = s

                if isinstance(s, six.string_types):
                    s = (s, True)
                elif s is True:
                    for var in get_variables(idx, elem):
                        s = sort_var_lex(var)
                        res.append((var, s))
                    continue
                elif callable(s):
                    s = (None, s)

                if not isinstance(s, tuple):
                    if not isinstance(s, collections.Iterable):
                        raise TypeError("spec %d has invalid format: %r" %
                                        (idx, elem))
                    s = (None, s)
                s = list(s)

                if len(s) != 2:
                    raise TypeError("spec %d has invalid format: %r" %
                                    (idx, elem))

                if s[0] is not None and s[0] not in get_variables(idx, elem):
                    raise ValueError(
                        "spec %d references non-existing variable: %r" %
                        (idx, s[0]))

                if s[1] is True:
                    s[1] = sort_var_lex(s[0])
                elif not callable(s[1]):
                    if not isinstance(s[1], collections.Iterable):
                        raise ValueError("spec %d has invalid format: %r" %
                                         (idx, elem))
                    if not isinstance(s[1], list):
                        s[1] = list(s[1])

                    if s[0] is None:
                        s[1] = sort_tick_from_list(list(s[1]))
                    else:
                        s[1] = sort_var_from_list(list(s[1]), s[0])
                else:
                    if s[0] is None:
                        s[1] = sort_instance_from_func(s[1])
                    else:
                        s[1] = sort_var_from_func(s[0], s[1])

                # check for repeated elements
                if s[0] in dict(res):
                    if s[0] is None:
                        raise ValueError(
                            "spec %d uses repeated non-variable format: %s" %
                            (idx, elem))
                    else:
                        raise ValueError(
                            "spec %d uses repeated variable name: %s" %
                            (idx, elem))

                res.append(s)

            return res
        spec = normalize(spec)

        # merged sorting function
        def sort_func(i1, i2):
            for elem in spec:
                res = elem[1](i1, i2)
                if res != 0:
                    return res
            return 0

        self._group.sort([], cmp=sort_func)

        if self.array is not None:
            old_indexes = list(self._indexes[i] for i in self._group)

        # update indexes
        def update_indexes(dim):
            for idx, instance in enumerate(dim._group):
                if instance in self._indexes:
                    dim._indexes[instance] = idx
        base = self.base if self.base is not None else self
        update_indexes(base)
        for view in base._views.values():
            update_indexes(view())

        if self.array is not None:
            self.array._sort(self, old_indexes)


class DimViewIndexes:
    """A view to the numeric indexes of a `Dim`.

    Examples
    --------
    >>> d = Dim("@entries@-@entry_size@",
    ...         ["1-8", None, "2-8", "2-16"])
    >>> list(d.indexes)
    [0, 1, 2, 3]
    >>> d.indexes[2]
    2
    >>> d.indexes["2-8"]
    2
    >>> d.indexes[1:]
    [1, 2, 3]
    >>> d.indexes[1:-1]
    [1, 2]
    >>> d.indexes[::"entries==2"]
    [2, 3]
    >>> d.indexes[[["1-8", 1], ["2-8", "1-8"]]]
    [[0, 1], [2, 0]]

    """
    def __init__(self, dimension, expand_slice=True):
        self.dimension = dimension
        self._expand_slice = expand_slice

    def __len__(self):
        return len(self.dimension)

    def __iter__(self):
        for i in self.dimension._group:
            yield self.dimension._indexes[i]

    def __getslice__(self, start, stop):
        if start is 0:
            start = None
        if stop == sys.maxint:
            return self[start::]
        else:
            return self[start:stop:]

    def __getitem__(self, index):
        indexes = self.dimension._get_indexes(index)

        if isinstance(indexes, slice) and self._expand_slice:
            return range(*indexes.indices(len(self.dimension)))
        else:
            if isinstance(indexes, numpy.ndarray):
                return indexes.tolist()
            else:
                return indexes


class DimViewTicks:
    """A view to the ticks of a `Dim`.

    Examples
    --------
    >>> d = Dim("@entries@-@entry_size@",
    ...         ["1-8", None, "2-8", "2-16"])
    >>> list(d.ticks)
    ['1-8', None, '2-8', '2-16']
    >>> d.ticks[1]
    >>> d.ticks[2]
    '2-8'
    >>> d.ticks["2-8"]
    '2-8'
    >>> d.ticks[1:]
    [None, '2-8', '2-16']
    >>> d.ticks[1:-1]
    [None, '2-8']
    >>> d.ticks[::"entries==2"]
    ['2-8', '2-16']
    >>> d.ticks[[["1-8", 1], ["2-8", "1-8"]]]
    [['1-8', None], ['2-8', '1-8']]

    """
    def __init__(self, dimension):
        self.dimension = dimension

    def __len__(self):
        return len(self.dimension)

    def _xlate(self, value):
        try:
            return self.dimension._expression.translate(value)
        except varref.XlateError:
            return None

    def __iter__(self):
        for i in self.dimension._group:
            yield self._xlate(i)

    def __getslice__(self, start, stop):
        if start is 0:
            start = None
        if stop == sys.maxint:
            return self[start::]
        else:
            return self[start:stop:]

    def __getitem__(self, index):
        indexes = self.dimension.indexes[index]

        def xlate(instance):
            if len(instance) == 0:
                return None
            else:
                return self._xlate(instance)

        assert not isinstance(indexes, slice)
        if numpy.shape(indexes) != tuple():
            if not isinstance(indexes, numpy.ndarray):
                indexes = numpy.array(indexes, dtype=object)
            for idx, elem in enumerate(indexes.flat):
                instance = self.dimension._group.get_index(elem)
                indexes.flat[idx] = xlate(instance)
            return indexes.tolist()

        else:
            assert isinstance(indexes, numbers.Integral)
            instance = self.dimension._group.get_index(indexes)
            return xlate(instance)


class DimViewInstances:
    """A view to the instances of a `Dim`.

    .. warning::

       You should not modify the instances returned by this object.

    Examples
    --------
    >>> d = Dim("@entries@-@entry_size@",
    ...         ["1-8", None, "2-8", "2-16"])
    >>> list(d.instances)
    [Instance({'entries': 1, 'entry_size': 8}), None, Instance({'entries': 2, 'entry_size': 8}), Instance({'entries': 2, 'entry_size': 16})]
    >>> d.instances[1]
    >>> d.instances[2]
    Instance({'entries': 2, 'entry_size': 8})
    >>> d.instances["2-8"]
    Instance({'entries': 2, 'entry_size': 8})
    >>> d.instances[1:]
    [None, Instance({'entries': 2, 'entry_size': 8}), Instance({'entries': 2, 'entry_size': 16})]
    >>> d.instances[1:-1]
    [None, Instance({'entries': 2, 'entry_size': 8})]
    >>> d.instances[::"entries==2"]
    [Instance({'entries': 2, 'entry_size': 8}), Instance({'entries': 2, 'entry_size': 16})]
    >>> d.indexes[d.instances[::"entries==2"]]
    [2, 3]
    >>> d.ticks[d.instances[::"entries==2"]]
    ['2-8', '2-16']
    >>> d.instances[[["1-8", 1], ["2-8", "1-8"]]]
    [[Instance({'entries': 1, 'entry_size': 8}), None], [Instance({'entries': 2, 'entry_size': 8}), Instance({'entries': 1, 'entry_size': 8})]]

    """
    def __init__(self, dimension):
        self.dimension = dimension

    def __len__(self):
        return len(self.dimension)

    def __iter__(self):
        for i in self.dimension._group:
            if len(i) == 0:
                yield None
            else:
                yield i

    def __getslice__(self, start, stop):
        if start is 0:
            start = None
        if stop == sys.maxint:
            return self[start::]
        else:
            return self[start:stop:]

    def __getitem__(self, index):
        indexes = self.dimension.indexes[index]

        assert not isinstance(indexes, slice)
        if numpy.shape(indexes) != tuple():
            if not isinstance(indexes, numpy.ndarray):
                indexes = numpy.array(indexes, dtype=object)
            for idx, elem in enumerate(indexes.flat):
                instance = self.dimension._group.get_index(elem)
                if len(instance) == 0:
                    instance = None
                indexes.flat[idx] = instance
            return indexes.tolist()

        else:
            assert isinstance(indexes, numbers.Integral)
            instance = self.dimension._group.get_index(indexes)
            if len(instance) == 0:
                instance = None
            return instance


class DimViewValues(object):
    """A view to the variable values of a `Dim`.

    This can be used as a shortcut to processing the elements of
    `DimViewInstances`. If more than one variable name is specified, values are
    returned as tuples.

    Examples
    --------
    The list of values of a given variable retains the order of the dimension:

    >>> d = Dim("@a@-@b@-@c@",
    ...         ["%d-%d-%d" % (a, b, c)
    ...          for a in range(3)
    ...          for b in range(3, 6)
    ...          for c in range(6, 9)])
    >>> d                               # doctest: +ELLIPSIS
    Dim('@a@-@b@-@c@',
        ['0-3-6',
         '0-3-7',
         '0-3-8',
         ...,
         '2-5-6',
         '2-5-7',
         '2-5-8'])
    >>> d.values["a"]                   # doctest: +ELLIPSIS
    [0, 0, 0, ..., 2, 2, 2]
    >>> d.unique_values["a"]
    OrderedSet([0, 1, 2])

    Multiple variables can also be specified:

    >>> d.values["a", "c"]              # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    [(0, 6), (0, 7), (0, 8),
     (0, 6), (0, 7), (0, 8),
     ...
     (2, 6), (2, 7), (2, 8),
     (2, 6), (2, 7), (2, 8)]
    >>> d.unique_values["a", "c"]
    OrderedSet([(0, 6),
                (0, 7),
                (0, 8),
                (1, 6),
                (1, 7),
                (1, 8),
                (2, 6),
                (2, 7),
                (2, 8)])

    Slices and regular indexes can also be used:

    >>> d.unique_values["b":-1]
    OrderedSet([(3,), (4,), (5,)])

    """
    def __init__(self, dimension, unique):
        self.dimension = dimension
        self.unique = unique

    def __getslice__(self, start, stop):
        if start is 0:
            start = None
        if stop == sys.maxint:
            return self[start::]
        else:
            return self[start:stop:]

    def __getitem__(self, index):
        variables = list(self.dimension.variables)

        def get_elem(elem, as_var):
            if isinstance(elem, six.string_types):
                try:
                    idx = variables.index(elem)
                except ValueError:
                    raise ValueError("unknown variable name %r" % elem)
                if not as_var:
                    elem = idx
            elif as_var:
                try:
                    elem = variables[elem]
                except ValueError:
                    raise ValueError("unknown variable name %r" % elem)
            return elem

        if isinstance(index, slice):
            slice_start = get_elem(index.start, False)
            slice_stop = get_elem(index.stop, False)
            index = slice(slice_start, slice_stop, index.step)
            index = tuple(range(*index.indices(len(variables))))

        if not isinstance(index, tuple):
            index = get_elem(index, True)
            if self.unique:
                res = sciexp2.common.utils.OrderedSet()
                for i in self.dimension._group:
                    res.add(i[index])
                return res
            else:
                res = []
                for i in self.dimension._group:
                    res.append(i[index])
                return res
        else:
            index = [get_elem(idx, True) for idx in index]
            if self.unique:
                res = sciexp2.common.utils.OrderedSet()
                for i in self.dimension._group:
                    res.add(tuple([i[idx] for idx in index]))
                return res
            else:
                res = []
                for i in self.dimension._group:
                    res.append(tuple([i[idx] for idx in index]))
                return res


__all__ = [
    "DimExpressionError", "DimExpression",
    "Dim",
    "DimViewIndexes", "DimViewTicks", "DimViewInstances", "DimViewValues",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Support for instances and associated abstractions."""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2008-2015, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import collections
import re
import six
from six.moves import cPickle as pickle
import weakref

from . import progress
from . import utils
from . import varref
from . import filter
from sciexp2.common import pp


_EXPORT = [set()]


class Instance (dict, pp.Pretty):
    """A map from variable names to values."""

    def __init__(self, source=None):
        """
        Parameters
        ----------
        source : :class:`dict` or ``None``
            Initial values for this instance.

        """
        if source is None:
            dict.__init__(self)
        else:
            dict.__init__(self, source)

    def __hash__(self):
        """Hashing function for adding into `InstanceGroup`."""
        return id(self)

    def __iter__(self):
        """Get an iterator to ``(key, value)`` pairs in the Instance."""
        if six.PY3:
            return iter(dict.items(self))
        else:
            return dict.iteritems(self)

    def copy(self):
        """Return a shallow copy of self."""
        return Instance(self)

    def deepcopy(self, *keys):
        """Return a copy of self, deepcopying the values of given keys."""
        res = Instance(self)
        for key in keys:
            res[key] = self[key].__class__(self[key])
        return res

    def _repr_pretty_(self, p, cycle):
        with self.pformat(p, cycle):
            p.pretty(dict(self))

    def __repr__(self):
        return pp.Pretty.__repr__(self)

    def __getstate__(self):
        progress.Stack.CURRENT()
        if len(_EXPORT[-1]) == 0:
            odict = self.__dict__
        else:
            odict = {}
            for k, v in six.iteritems(self):
                if k in _EXPORT[-1]:
                    odict[k] = v
        return odict

    def __setstate__(self, contents):
        progress.Stack.CURRENT()
        self.__dict__.update(contents)


class InstanceExtractor:
    """Extract instances out from plain text.

    Automates the creation of instances from various strings, following a given
    template.

    Examples
    --------
    You can use some pre-defined variable-referencing patterns to extract the
    instances:

    >>> e = InstanceExtractor('superfluous @var1@ text', "@")
    >>> e.extract('superfluous (or not) text')
    Instance({'var1': '(or not)'})

    You can explicitly use `named groups <re>` in your template expression
    in order to support more complex expressions:

    >>> e = InstanceExtractor('superfluous @var1@ text [^0-9]*(?P<number>[0-9]+)?.*', "@")
    >>> e.extract('superfluous (or not) text and a trail')
    Instance({'number': None, 'var1': '(or not)'})
    >>> e.extract('superfluous (or not) text with some 123 some numbers')
    Instance({'number': '123', 'var1': '(or not)'})

    """

    def __init__(self, template, varre):
        """
        Parameters
        ----------
        template : string
            Reference regexp with variables for extracting values.
        varre
            A valid argument to `~sciexp2.common.varref.build_varre`.

        """
        self._cre = re.compile(varref.expr_to_regexp(varre, template))

    def extract(self, source):
        """Extract an Instance from the given source text.

        Returns
        -------
        Instance or None :
            The extracted `Instance`, or ``None`` if `source` cannot be
            expanded according to the reference `template`.

        """
        match = self._cre.match(source)
        try:
            groups = match.groupdict()
        except AttributeError:
            return None
        return Instance(groups)


class ViewError (Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class InstanceGroup (pp.Pretty):
    """A group of `Instance` objects suitable for searching.

    Provides a fast way to retrieve a set of Instance objects that have certain
    variables set or certain variable values.

    Parameters
    ----------
    instances : iterable of Instance objects, optional
        Initial contents. Contents are copied during initialization.
    cache, cache_proactive, cache_others
        Arguments to `cache_set`.
    view_able : bool, optional
        Whether to enable the use of `view`.

    See also
    --------
    cache_set

    Notes
    -----
    .. warning::

       You should **not** modify the results of accessing an InstanceGroup,
       unless you know what you're doing.

       As this operation returns the same data being used internally, changes
       to it must be kept consistent with the cache (see `cache_reset`).

    .. warning::

       You should **not** add or extend a group with instances that have the
       same contents as other instances already present in the group.

       .. todo:: An `InstanceGroup` should mimic an `OrderedSet`

    Examples
    --------
    >>> g = InstanceGroup([{'a':4, 'b':3}, {'a':2, 'b':1}, {'a':4}])
    >>> g
    InstanceGroup([Instance({'a': 4, 'b': 3}),
                   Instance({'a': 2, 'b': 1}),
                   Instance({'a': 4})])

    Testing the existence of variables:

    >>> 'a' in g
    True
    >>> 'c' in g
    False

    Getting all the possible values for a variable:

    >>> g['a']                          # doctest: +NORMALIZE_WHITESPACE
    OrderedDict([(4, OrderedSet([Instance({'a': 4, 'b': 3}), Instance({'a': 4})])),
                 (2, OrderedSet([Instance({'a': 2, 'b': 1})]))])
    >>> g['b']                          # doctest: +NORMALIZE_WHITESPACE
    OrderedDict([(3, OrderedSet([Instance({'a': 4, 'b': 3})])),
                 (1, OrderedSet([Instance({'a': 2, 'b': 1})]))])

    Getting all the instances with a specific value:

    >>> g['a'][4]                          # doctest: +NORMALIZE_WHITESPACE
    OrderedSet([Instance({'a': 4, 'b': 3}), Instance({'a': 4})])

    Extending a group with another:

    >>> h = InstanceGroup([{'a':1}, {'a':2}])
    >>> h
    InstanceGroup([Instance({'a': 1}), Instance({'a': 2})])
    >>> g + h
    InstanceGroup([Instance({'a': 4, 'b': 3}),
                   Instance({'a': 2, 'b': 1}),
                   Instance({'a': 4}),
                   Instance({'a': 1}),
                   Instance({'a': 2})])
    >>> g += h
    >>> g
    InstanceGroup([Instance({'a': 4, 'b': 3}),
                   Instance({'a': 2, 'b': 1}),
                   Instance({'a': 4}),
                   Instance({'a': 1}),
                   Instance({'a': 2})])
    >>> g['a']                          # doctest: +NORMALIZE_WHITESPACE
    OrderedDict([(4, OrderedSet([Instance({'a': 4, 'b': 3}), Instance({'a': 4})])),
                 (2, OrderedSet([Instance({'a': 2, 'b': 1}), Instance({'a': 2})])),
                 (1, OrderedSet([Instance({'a': 1})]))])

    """

    def __init__(self, instances=None, cache=True,
                 cache_proactive=False, cache_others=None,
                 view_able=False):
        self._view_able = view_able
        self._base = None
        self._views = {}
        self._instances = utils.OrderedSet(view_able=self._view_able)
        self._reverse = dict()

        self._cache = True
        self._cache_proactive = False
        self._cache_others = None
        self.cache_set(cache, cache_proactive, cache_others)

        if instances is not None:
            for i in instances:
                self.add(Instance(i))

    def view(self, index):
        assert self._view_able
        assert isinstance(index, slice)
        res = InstanceGroup([], cache=self._cache,
                            cache_proactive=self._cache_proactive,
                            cache_others=self._cache_others, view_able=True)
        res._instances = self._instances.view(index)
        res._reverse = dict(self._reverse)
        res.cache_reset()
        res.add = self._add_view
        self.add = self._add_view
        res._base = self
        self._views[id(self)] = weakref.ref(res)
        return res

    def copy(self):
        """Return a copy."""
        return InstanceGroup(self,
                             cache=self._cache,
                             cache_proactive=self._cache_proactive,
                             cache_others=self._cache_others,
                             view_able=self._view_able)

    def get_index(self, index):
        """Get instance at the  `index`'th position."""
        return self._instances.get_index(index)

    def add(self, instance):
        """Add an `Instance` into this group.

        The caching state will have an impact on the performance of this
        operation.

        See also
        --------
        cache_set

        """
        # actual function set by `_set_cache_proactive`
        assert False

    def _set_cache_proactive(self, proactive, others=None):
        if isinstance(proactive, bool):
            if self._cache_proactive:
                self.add = self._add_cache_all
            else:
                self.add = self._add_cache_none

        else:
            if len(proactive) == 0 and others is None:
                self.add = self._add_cache_direct
            elif others is None:
                proactive = set(proactive)
                self.add = self._add_cache_list
            else:
                for var in others:
                    self._cache_reset(var)
                proactive = tuple(proactive)
                if len(proactive) == 0:
                    self.add = self._add_cache_direct
                else:
                    self.add = self._add_cache_list_only

            prefetch = [var for var in proactive
                        if var in set(self.variables())]
            self.cache_prefetch(prefetch, skip_cached=False)

        self._cache_proactive = proactive
        self._cache_others = others

    def _add_view(self, instance):
        raise ViewError("cannot add to a view")

    def _add_cache_direct(self, instance):
        assert isinstance(instance, Instance)
        self._instances.add(instance)

    def _add_cache_all(self, instance):
        assert isinstance(instance, Instance)
        self._instances.add(instance)
        for k, v in instance:
            self._cache_add(k, v, instance)

    def _add_cache_none(self, instance):
        assert isinstance(instance, Instance)
        self._instances.add(instance)
        # just initialize existence of variables
        for k, _ in instance:
            self._cache_reset(k)

    def _add_cache_list(self, instance):
        assert isinstance(instance, Instance)
        self._instances.add(instance)
        for k, v in instance:
            if k in self._cache_proactive:
                self._cache_add(k, v, instance)
            else:
                self._cache_reset(k)

    def _add_cache_list_only(self, instance):
        assert isinstance(instance, Instance)
        self._instances.add(instance)
        for k in self._cache_proactive:
            self._cache_add(k, instance[k], instance)

    def __iadd__(self, obj):
        """Extend this group with one more instance or group.

        .. todo::

           Should probably delete extension with instances.

        """
        try:
            self.add(obj)
        except Exception:
            if not isinstance(obj, Instance):
                self.extend(obj)
            else:
                raise
        return self

    def extend(self, instances):
        """Extend this group with the contents of another group."""
        assert isinstance(instances, InstanceGroup)
        for instance in instances:
            self.add(instance)

    def __add__(self, instances):
        """x.__add__(y) <==> x + y

        Equivalent to a new `InstanceGroup` extended with both groups.

        """
        res = InstanceGroup(self)
        res.extend(instances)
        return res

    def __contains__(self, variable):
        """Syntax sugar for `has_variable`."""
        return variable in self._reverse

    def has_variable(self, variable):
        """Whether the group contains any Instance with given variable name."""
        return variable in self._reverse

    def has_instance(self, instance):
        """Whether this group contains the given Instance."""
        return instance in self._instances

    def variables(self):
        """The sequence of variables in this group."""
        return six.iterkeys(self._reverse)

    def __getitem__(self, var):
        """Get all values for the given variable."""
        res = self._reverse[var]
        if res is None:  # JIT caching
            res = collections.OrderedDict()
            for instance in self._instances:
                try:
                    val = instance[var]
                except KeyError:
                    continue
                try:
                    vals = res[val]
                except KeyError:
                    vals = utils.OrderedSet()
                    res[val] = vals
                vals.add(instance)
            if self._cache:
                self._reverse[var] = res
        return res

    def _cache_add(self, key, value, instance):
        try:
            keys = self._reverse[key]
        except KeyError:
            keys = collections.OrderedDict()
            self._reverse[key] = keys

        try:
            vals = keys[value]
        except TypeError:               # 'key' uncached
            keys = collections.OrderedDict()
            self._reverse[key] = keys
            vals = utils.OrderedSet()
            keys[value] = vals
        except KeyError:                # 'value' not present
            vals = utils.OrderedSet()
            keys[value] = vals

        vals.add(instance)

    def _cache_reset(self, key):
        self._reverse[key] = None

    def cache_get(self):
        """Get the caching state."""
        return self._cache

    def cache_set(self, enable, proactive=None, others=None):
        """Set the caching status for this group.

        Caching offers greater speed of getting a specific variable multiple
        times (e.g., ``g['a']``), at the expense of a higher memory footprint.

        Parameters
        ----------
        enable : bool
           Caching status. When the caching is disabled, automatically calls
           `cache_reset`.
        proactive : bool or sequence of variable names, optional
           Proactively cache slected variables when adding new instances
           (default is to keep last value).
        others : sequence of variable names, optional
           List of other variables that the group does/will contain (default is
           to keep last value).

        See also
        --------
        cache_get, cache_reset, cache_prefetch

        Notes
        -----
        If caching is disabled, every call to `add` will reset the cache for
        the variables of the added instance.

        If `proactive` is `True`, it will proactively cache all variables, if
        it is `False` it will not. If it is a sequence of variable names, only
        these variables will be proactively cached.

        You can also provide an empty `proactive` list and an appropriate
        `others` list to get zero-caching during instance addition, which is
        the fastest option available.

        .. warning ::

           Providing `others` will greatly enhance the performance of `add`,
           but the contents will be inconsistent if `others` does not contain
           exactly all the variables of the instances previously present or
           added before interacting with the group.

           You can use `cache_reset` with the `hard` argument to ensure
           consistency.

        """
        assert isinstance(enable, bool)
        if proactive is not None:
            if enable is False:
                raise ValueError("Cannot use proactive caching when "
                                 "caching is disabled")
            if not isinstance(proactive, bool) and \
               (
                   not isinstance(proactive, collections.Iterable) or
                   not all([isinstance(elem, str) for elem in proactive])
               ):
                raise TypeError("Invalid `proactive` caching "
                                "value: %s" % proactive)
        if others is not None:
            if proactive is None:
                raise ValueError(
                    "Cannot use 'others' without proactive caching")
            if not isinstance(others, bool) and \
               (
                   not isinstance(others, collections.Iterable) or
                   not all([isinstance(elem, str) for elem in others])
               ):
                raise TypeError("Invalid `others` caching "
                                "value: %s" % others)

        if self._cache != enable:
            self._cache = enable
            if not enable:
                self.cache_reset()

        if proactive is not None:
            self._set_cache_proactive(proactive, others)

    def cache_reset(self, hard=False):
        """Reset the cache of this group.

        Parameters
        ----------
        hard : bool optional
            Perform a costly "hard" cache reset by inspecting all instances in
            the group.

        See also
        --------
        cache_get, cache_set

        """
        if hard:
            self._reverse = dict()
            for instance in self._instances:
                for var in six.iterkeys(instance):
                    self._cache_reset(var)
        else:
            for var in six.iterkeys(self._reverse):
                self._cache_reset(var)

    def cache_prefetch(self, variables=None, skip_cached=True):
        """Enable caching and force it for the given variables.

        Parameters
        ----------
        variables : sequence of variable names, optional
            Variables to prefetch. Uses all variables when ``None``.
        skip_cached : bool, optional
            Do not recompute caching if a variable is already cached.

        See also
        --------
        cache_get, cache_set

        Notes
        -----
        When prefetching a variable that is already cached, not necessarily all
        its values will be present on the cache (only those previously
        referenced); thus the `skip_cached` argument.

        """
        self._cache = True

        # select variables
        if variables is None:
            variables = self.variables()
        else:
            self_variables = list(self.variables())
            for var in variables:
                if var not in self_variables:
                    raise ValueError("No such variable: '%s'" % var)

        # uncache variables to be cached
        variables = set(variables)
        for var in set(variables):
            if self._reverse[var] is not None and skip_cached:
                variables.remove(var)
            else:
                self._cache_reset(var)

        # start caching
        if len(variables) == 0:
            return
        for instance in self._instances:
            for var in variables:
                if var not in instance:
                    continue
                self._cache_add(var, instance[var], instance)

    def __len__(self):
        """The number of instances in this group."""
        return len(self._instances)

    def __iter__(self):
        """An iterator to the instances in the group."""
        return iter(self._instances)

    def select(self, filter_, allow_unknown=False):
        """Generate a sequence of instances matching the given filter.

        Parameters
        ----------
        filter_ : filter
            Filter establishing which instances to select from this group
            (accepts anything convertible into a filter).
        allow_unknown : bool or callable, optional
            Whether to allow the filter to reference variables not present in
            all instances of the group (in which case the instance is not
            selected). If a callable is used, it will receive the instance and
            its result will be used.

        """
        filter_ = filter.Filter(filter_)

        filter_dict = filter_.as_dict()
        if self._cache and filter_dict is not None:
            if len(filter_dict) == 0:
                return iter(self._instances)

            # optimize through set intersection
            cache_old = self.cache_get()
            self.cache_prefetch(variables=list(filter_dict.keys()))

            instances = None
            for var, val in six.iteritems(filter_dict):
                # cache_prefetch ensures variables are present
                var_values = self[var]
                try:
                    self_instances = var_values[val]
                except KeyError:
                    instances = utils.OrderedSet()
                    break
                if instances is None:
                    instances = self_instances
                else:
                    instances = instances & self_instances
            res = iter(instances)

            self.cache_set(cache_old)
            return res

        else:
            def generate(instances):
                for instance in instances:
                    try:
                        match = filter_.match(instance)
                    except filter.FilterError:
                        if callable(allow_unknown):
                            if not allow_unknown(instance):
                                raise
                        elif not bool(allow_unknown):
                            raise
                    else:
                        if match:
                            yield instance
            return generate(self._instances)

    def sorted(self, *args, **kwargs):
        """Same as `sort`, but returns a sorted copy."""
        res = self.copy()
        res.sort(*args, **kwargs)
        return res

    def sort(self, variables, cmp=None):
        """Sort group in-place by the given variables.

        Parameters
        ----------
        variables : list of strings
            Variable names for which to sort their values. The instances are
            sorted according to the order of the given variable names.
        cmp : func, optional
            Custom comparison function of two arguments (instances). Returns
            -1, 0, 1 if the first argument is considered less than, equal to or
            greater than the second argument (respectively).

        """
        # get list of all variables in group
        variables = list(variables)
        for var in self.variables():
            if var not in variables:
                variables.append(var)

        # transform values into a list, sorted according to `variables`
        def to_list(instance):
            return [instance[var] for var in variables]

        if cmp is None:
            kwargs = {"key": to_list}
        else:
            kwargs = {"cmp": cmp}

        self._instances.sort(**kwargs)

    def _repr_pretty_(self, p, cycle):
        with self.pformat(p, cycle):
            p.pretty(list(self._instances))

    #######################################################################
    # pickling
    #######################################################################

    def dump(self, file_obj, export=set()):
        """Write a pickled representation.

        Parameters
        ----------
        file_obj : file
            The destination file.
        export : set of strings, optional
            Variables that must be exported. Defaults to all.

        See also
        --------
        pickle

        """
        _EXPORT.append(set(export))
        start = progress.get_pickle(self)
        stop = start.get_stop()
        pickle.dump((start, self, stop), file_obj, -1)
        _EXPORT.pop()

    def __getstate__(self):
        odict = self.__dict__.copy()
        new_reverse = dict()
        for key in odict["_reverse"]:
            if len(_EXPORT[-1]) == 0 or key in _EXPORT[-1]:
                new_reverse[key] = None
        odict["_reverse"] = new_reverse
        if "add" in odict:
            del odict["add"]
        del odict["_base"]
        del odict["_views"]
        return odict

    @staticmethod
    def load(file_obj):
        """Load a pickled representation.

        See also
        --------
        pickle

        """
        _, res, _ = pickle.load(file_obj)
        assert isinstance(res, InstanceGroup)
        return res

    def __setstate__(self, odict):
        self.__dict__.update(odict)
        self._base = None
        self._views = {}
        self._set_cache_proactive(self._cache_proactive, self._cache_others)


class ExpanderResult:
    """The results of an expansion.

    Parameters
    ----------
    text : string
        The expanded text.
    grouped : bool, optional
        Whether the `instances` attribute is an
        `~sciexp2.common.utils.OrderedSet` or an `InstanceGroup`.
    with_filter : bool, optional
        Whether the `filter` attribute has to contain a
        `~sciexp2.common.filter.Filter` grouping all contained instances
        together or ``None``.

    Attributes
    ----------
    text
        The expanded text.
    instances
        `Instance` objects with the same expansion.
    filter
        `~sciexp2.common.filter.Filter` grouping all contained
        instances together.

    """

    def __init__(self, text, grouped=False, with_filter=False):
        self.text = text
        if grouped:
            self.instances = InstanceGroup()
        else:
            self.instances = utils.OrderedSet()
        if with_filter:
            self.filter = filter.Filter()
        else:
            self.filter = None

    def __iadd__(self, instance):
        """Syntax sugar for `add`.

        .. todo::

           Should probably delete.

        """
        self.add(instance)
        return self

    def add(self, instance):
        """Add a new Instance into the result."""
        self.instances.add(instance)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self.text))


class Expander:
    """Expand expressions with all instances in an `InstanceGroup`.

    Parameters
    ----------
    varre
        A valid argument to `~sciexp2.common.varref.build_varre`.

    Examples
    --------
    Given a template expression, generates one translated string for each
    `Instance` in the given `InstanceGroup`:

    >>> e = Expander('@')
    >>> g = InstanceGroup([{'a':4, 'b':3}, {'a':2, 'b':1}, \
                           {'a':4, 'b':3, 'c':2}])
    >>> r = e.expand('@a@ -> @b@', g, with_filter = True)
    >>> r
    OrderedDict([('4 -> 3', ExpanderResult('4 -> 3')), ('2 -> 1', ExpanderResult('2 -> 1'))])
    >>> r['2 -> 1'].instances
    OrderedSet([Instance({'a': 2, 'b': 1})])
    >>> r['4 -> 3'].instances
    OrderedSet([Instance({'a': 4, 'b': 3}), Instance({'a': 4, 'b': 3, 'c': 2})])
    >>> r['4 -> 3'].filter
    Filter("a == 4 && b == 3")
    >>> r = e.expand('@a@ -> @b@', g, grouped = True)
    >>> r['4 -> 3'].instances
    InstanceGroup([Instance({'a': 4, 'b': 3}), Instance({'a': 4, 'b': 3, 'c': 2})])

    """

    def __init__(self, varre):
        self._xlator = varref.DictXlator(varre)
        self._tokenizer = varref.Tokenizer(varre)

    def expand(self, source, group, filter_=None, allow_unknown=False,
               grouped=False, with_filter=False, recursive=False):
        """Expand a source text given a reference InstanceGroup.

        Parameters
        ----------
        source : string
            The source text with variable references to expand.
        group : InstanceGroup
            `InstanceGroup` to expand from.
        filter_ : Filter instance or filter string, optional
            `~sciexp2.common.filter.Filter` to select instances for the
            expansion. Instances in *group* not matching *filter_* will not be
            used to expand on *source*.
        allow_unknown : bool or callable, optional
            Whether to allow the filter to reference variables not present in
            all instances of the group (in which case the instance is not
            selected for expansion). If a callable is used, it will receive the
            instance and its result will be used.
        grouped : bool, optional
            Whether to return expansions inside an
            `~sciexp2.common.utils.OrderedSet` (``False``) or an
            `InstanceGroup` (``True``).
        with_filter : bool, optional
            Whether to set `ExpanderResult.filter`, or ``None`` otherwise.
        recursive : bool, optional
            Whether to recursively translate values in *group* on each
            expansion.

        Returns
        -------
        OrderedDict
            Mapping from translated text to an `ExpanderResult` instance.

        Notes
        -----
        .. todo::

           Would be better to provide an extra *filter* method instead of using
           this method with so verbose arguments and a *source* with no
           variable references.

        .. todo::

           Incorrect with recursive translations.

        """
        if filter_ is None:
            filter_ = filter.Filter()
        text = str(source)

        assert isinstance(group, InstanceGroup)
        assert isinstance(filter_, filter.Filter)
        filter_.validate(group)

        variables = list(self._tokenizer.vars(text))

        res = collections.OrderedDict()

        # shortcut for an empty InstanceGroup
        if len(group) is 0:
            res[text] = ExpanderResult(text, grouped, filter_)
            return res

        # shortcut for non-expandable source
        if len(variables) is 0:
            result = ExpanderResult(text, grouped, filter_)
            for instance in group:
                try:
                    match = filter_.match(instance)
                except filter.FilterError:
                    if callable(allow_unknown):
                        if not allow_unknown(instance):
                            raise
                    elif not bool(allow_unknown):
                        raise
                else:
                    if match:
                        result += instance
            res[text] = result
            return res

        xlator = self._xlator.xlate if not recursive \
                 else self._xlator.xlate_rec

        if with_filter:
            def _with_filter(instance):
                for var in variables:
                    # TODO: Incorrect with recursive translations.
                    val = instance[var]
                    if isinstance(val, six.string_types):
                        fmt = "%s == '%s'"
                    else:
                        fmt = "%s == %s"
                    result.filter &= filter.Filter(fmt % (var, val))
        else:
            def _with_filter(instance):
                pass

        for instance in group:
            if not filter_.match(instance):
                continue

            xlated = xlator(text, instance, allow_unxlated=not recursive)

            try:
                result = res[xlated]
            except KeyError:
                result = ExpanderResult(xlated, grouped, filter_)
                _with_filter(instance)
                res[xlated] = result

            result += instance

        return res

    def flatten(self, source, group):
        """Flatten all variable references in the source text.

        Recursively translates with `group` all variable references in `source`
        until (but not including) they contain immediate values (not variable
        references) or variables are multi-evaluated.

        Examples
        --------
        >>> e = Expander("@")
        >>> g = InstanceGroup([{'a':4, 'b':'foo', 'c':'@a@-@b@'}])
        >>> e.flatten('bar: @c@', g)
        'bar: @a@-@b@'

        """
        was_cached = group.cache_get()
        group.cache_prefetch()

        res = self._flatten(source, group)

        group.cache_set(was_cached)

        return res

    def _flatten(self, source, group):
        var_fmt = self._tokenizer.var_start + "%s" + self._tokenizer.var_end
        res = ""
        for is_var, tok in self._tokenizer.tokenize(source):
            if is_var:
                if len(group[tok]) is 1:  # uniquely evaluated
                    value = next(six.iterkeys(group[tok]))
                    if len(list(self._tokenizer.vars(value))) is 0:
                        # has immediate value
                        res += var_fmt % tok
                    else:
                        # flattened value
                        res += self._flatten(value, group)
                else:  # multiple values
                    res += var_fmt % tok
            else:  # literal
                res += tok
        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import absolute_import

import os
import sys
from distutils.core import setup
from distutils.cmd import Command

from distutils.dir_util import remove_tree
from distutils import log
from distutils.command.clean import clean


version_file = os.path.join("sciexp2", "version.py")
with open(version_file) as f:
    code = compile(f.read(), version_file, 'exec')
    exec(code) #, global_vars, local_vars)


# build sphinx docs
sphinx_base = "doc"
sphinx_build = os.path.join(sphinx_base, "_build")
sphinx_ref = os.path.join(sphinx_base, "_reference")

# commands specialized for sphinx
try:
    from sphinx.setup_command import BuildDoc
except:
    BuildDoc = None


class Clean (clean):
    def run(self):
        clean.run(self)
        if self.all:
            for directory in (sphinx_build, sphinx_ref):
                if os.path.exists(directory):
                    print(directory, self.dry_run)
                    remove_tree(directory, dry_run=self.dry_run)
                else:
                    log.debug("'%s' does not exist -- can't clean it",
                              directory)


# run tests
class Test (Command):
    description = "run unit tests"
    user_options = [
        ('test-without-doctest', None, 'test documentation strings'),
        ('test-with-coverage', None, 'provide a coverage report'),
        ('test-with-profile', None, 'profile code run during tests'),
    ]

    def initialize_options(self):
        self.test_without_doctest = False
        self.test_with_coverage = False
        self.test_with_profile = False

    def finalize_options(self):
        pass

    def run(self):
        import sciexp2.tests
        res = sciexp2.tests.run(
            with_doctest=not self.test_without_doctest,
            with_coverage=self.test_with_coverage,
            with_profile=self.test_with_profile,
        )
        if not res.success:
            sys.exit(1)

opts = dict(name=NAME,
            version=VERSION,
            description=DESCRIPTION,
            long_description=LONG_DESCRIPTION,
            author=AUTHOR,
            author_email=AUTHOR_EMAIL,
            maintainer=MAINTAINER,
            maintainer_email=MAINTAINER_EMAIL,
            url=URL,
            license=LICENSE,
            classifiers=CLASSIFIERS,
            packages=PACKAGES,
            package_data=PACKAGE_DATA,
            scripts=SCRIPTS,
            requires=REQUIRES,
            install_requires=INSTALL_REQUIRES,
            provides=PROVIDES,
            platforms="any",

            cmdclass={
                'clean': Clean,
                'test': Test,
            },
            command_options={
                'build_sphinx': {
                    'source_dir': ('setup.py', sphinx_base),
                    'build_dir':  ('setup.py', sphinx_build),
                    'all_files':  ('setup.py', True),
                    'version': ('setup.py', VERSION),
                    'release': ('setup.py', VERSION),
                },
            })

if BuildDoc is not None:
    opts["cmdclass"]["build_sphinx"] = BuildDoc

if __name__ == "__main__":
    setup(**opts)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Detect and manipulate references to variables.

The variable syntax is defined by the arguments to `build_varre`.

.. autodata:: REGEXPS

"""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2008-2013, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import abc
import collections
import functools
import re
import six

from . import utils


# Base functionality

#: Dictionary mapping the "name" of a variable referencing syntax, to the
#: definition of how to detect such references.
#:
#: The values on the dictionary must contain a 3-element tuple:
#:  * Prefix enclosing the variable name.
#:  * Regular expression for the variable name.
#:  * Postfix enclosing the variable name.
REGEXPS = {
    # <variable start str>, <variable name regexp>, <variable end str>
    "@": ("@", "[^@]+", "@"),
    "$": ("$(", "[^)]+", ")"),
    }


def _parts2varre(parts, ignore=None):
    """Build a regexp to match variable names."""
    if len(parts) != 3:
        raise ValueError("Variable regexp with invalid number of " +
                         "parts: " + str(parts))
    prefix = "".join(["\\" + char for char in parts[0]])
    prefix = prefix + "(?<!\\\\" + prefix + ")"
    varname = parts[1]
    postfix = "".join(["\\" + char for char in parts[2]])
    postfix = postfix + "(?<!" + postfix[:-1] + "\\\\" + postfix[-1:] + ")"
    if ignore is None:
        return prefix + "(" + varname + ")" + postfix
    else:
        assert isinstance(ignore, collections.Iterable) and \
            not isinstance(ignore, six.string_types)
        # negative lookbehind of ignores
        ignore_behind = ["(?<!%s%s)" % (prefix, elem) for elem in ignore]
        ignore_behind = "".join(ignore_behind)
        # negative lookahead of ignores
        ignore_ahead = "(?!" + (postfix + "|").join(ignore) + postfix + ")"
        # <not preceeded by ignores><prefix>(<not an ignore><varname>)<postfix>
        return ignore_behind + prefix + \
            "(" + ignore_ahead + varname + ")" + \
            postfix

_CREGEXPS = dict([(key, re.compile(_parts2varre(val)))
                  for key, val in REGEXPS.items()])


def build_varre(name, ignore=None):
    """Build a compiled regexp for intentifying variables.

    Parameters
    ----------
    name : str or tuple
        Either a key in :data:`REGEXPS` for a pre-existing named variable
        syntax, or the corresponding 3-element tuple.
    ignore : list(str), optional
        Sequence of variable names that must be ignored when searching for
        variable references.

    Returns
    -------
    tuple
        A 2-element tuple with the 3-element tuple (useful in case `name`
        referenced a key in :data:`REGEXPS`) and the compiled regexp for
        identifying such variable references.
    """
    varre = None
    cvarre = None
    if name in REGEXPS:
        parts = REGEXPS[name]
        if ignore is None:
            cvarre = _CREGEXPS[name]
        else:
            cvarre = re.compile(_parts2varre(parts, ignore))
    else:
        parts = name
        varre = _parts2varre(parts, ignore)
        cvarre = re.compile(varre)
    return parts, cvarre


def expr_to_regexp(varre, expr, contents=".+"):
    """Convert variable references into a regular expression with named groups.

    Parameters
    ----------
    varre
        A valid argument to `build_varre`.
    expr : str
        Expression to transform.
    contents : str, optional
        Regular expression for named group contents.

    """
    variables = set()
    res = ""
    for isvar, text in Tokenizer(varre).tokenize(expr):
        if isvar:
            if text in variables:
                res += "(?P=%s)" % text
            else:
                res += "(?P<%s>%s)" % (text, contents)
                variables.add(text)
        else:
            res += text
    return res


def regexp_get_names(regexp):
    """Get a list of the group names in a regular expression."""
    cre = re.compile(regexp)
    return [k for v, k in sorted(
        [(v, k)
         for k, v in six.iteritems(cre.groupindex)]
    )]


def expr_get_vars(varre, expr, accept_regexp=False):
    """Get a list with the variable names present in given expression.

    Parameters
    ----------
    varre
        A valid argument to `build_varre`.
    expr : str
        Expression to transform.
    accept_regexp : bool, optional
        Whether to accept a mix of variable references and regexp named groups.

    """
    if accept_regexp:
        return regexp_get_names(expr_to_regexp(varre, expr))
    else:
        res = []
        variables = set()
        for var in Tokenizer(varre).vars(expr):
            if var not in variables:
                variables.add(var)
                res.append(var)
        return res


class Tokenizer (six.with_metaclass(utils.SelectiveClassMemoize)):
    """Split a string at variable reference boundaries."""

    def __init__(self, varre, ignore=None):
        """
        Parameters
        ----------
        varre
            A valid argument to `build_varre`.
        ignore : list(str), optional
            Variable names to ignore during tokenization.
        """
        (self.var_start, _, self.var_end), self._cre = build_varre(
            varre, ignore)

    def tokenize(self, source):
        """Generator to tokenize a string.

        Generates a sequence of ``(is_variable, token_text)`` pairs.

        The tokenized variables names are stripped off of their prefix and
        postfix variable reference syntax.

        Notes
        -----
        `source` is automatically stringified to provide transparency with
        non-string inputs.
        """
        if not isinstance(source, six.string_types):
            text = str(source)
        else:
            text = source
        tokens = self._cre.split(text)
        if len(tokens) > 0 and tokens[0] == "":
            isvar = True
            tokens = tokens[1:]
        else:
            isvar = False
        for tok in tokens:
            yield (isvar, tok)
            isvar = not isvar

    def vars(self, source):
        """Generator to extract only variable names from a string."""
        for tok in self.tokenize(source):
            if tok[0]:
                yield tok[1]

Tokenizer.memoize_new("@")
Tokenizer.memoize_new("@", None)
Tokenizer.memoize_new("@", set())
Tokenizer.memoize_new("$")
Tokenizer.memoize_new("$", None)
Tokenizer.memoize_new("$", set())


# Variable translators

class XlateError (Exception):
    """Reference to unknown variable."""
    def __init__(self, variables, source, current, subst):
        self.variables = variables
        self.source = source
        self.current = current
        self.subst = subst

    def __str__(self):
        msg = ["Missing translation(s) for: %s" % ", ".join(self.variables),
               "Original source text: '%s'" % self.source]
        if self.current is not None:
            msg.append("Current text: '%s'" % self.current)
        if self.subst is not None:
            msg.append("Offending substitution: %s" % self.subst)
        return "\n".join(msg)


class RecursiveError (Exception):
    """Recursive variable translation."""
    def __init__(self, variables, source):
        self.variables = variables
        self.source = source

    def __str__(self):
        msg = ["Possible recursive translation for: %s" %
               ", ".join(self.variables),
               "Original source text: '%s'" % self.source]
        return "\n".join(msg)


class _XlatorMetaclass (utils.SelectiveClassMemoize, abc.ABCMeta):
    pass


class Xlator (six.with_metaclass(_XlatorMetaclass)):
    """Generic class for translating references to variables into values."""

    def __init__(self, varre, ignore=None):
        """
        Parameters
        ----------
        varre
            A valid argument to `build_varre`.
        ignore : list(str), optional
            Variable names to ignore during translation.
        """
        (self.var_start, _, self.var_end), self._cre = build_varre(
            varre, ignore)
        self._varre = varre

    @abc.abstractmethod
    def _xlate(self, source, subst):
        pass

    def xlate(self, source, subst=None, allow_unxlated=False):
        """Translate a string with the given dict.

        Parameters
        ----------
        source : str
            String with variable references.
        subst
            The definition of translations to apply.
            The interpretation of this argument depends on the specific
            subclass.
        allow_unxlated : bool, optional
            Allow for some variable references to remain untranslated if they
            are not in `subst` (otherwise raise an exception).

        Returns
        -------
        str
            String with all variable references replaced with the contents of
            `subst`.

        Notes
        -----
        If `source` is not a string, simply returns it with no translations
        applied.
        """
        # TODO: this is extremely hackish, as we account for the "special"
        #       value in PrepDictXlator
        if source is not None and not isinstance(source, str):
            return source

        text = source
        res, unxlated = self._xlate(text, subst)

        if not allow_unxlated and unxlated is not None:
            raise XlateError(unxlated, source, None, subst)
        return res

    def xlate_rec(self, source, subst=None, allow_unxlated=False):
        """Recursively apply `xlate`."""
        if not isinstance(source, six.string_types):
            return source

        tries = 0
        text = source
        prev = None
        while prev != text:
            prev = text
            text, unxlated = self._xlate(text, subst)

            # For the sake of performance, do not check real recursion but
            # instead look for a "long enough" loop
            if tries > 100:
                tokenizer = Tokenizer(self._varre)
                variables = list(tokenizer.vars(text))
                raise RecursiveError(variables, source)
            tries += 1

        if not allow_unxlated and unxlated is not None:
            raise XlateError(unxlated, source, prev, subst)
        return text


class NullXlator (Xlator):
    """Eliminate variable references from a string.

    Argument `subst` in `xlate` must be ``None``.
    """
    def _xlate(self, source, subst=None):
        assert subst is None
        return self._cre.sub("", str(source)), None

NullXlator.memoize_new("@")
NullXlator.memoize_new("@", None)
NullXlator.memoize_new("@", set())
NullXlator.memoize_new("$")
NullXlator.memoize_new("$", None)
NullXlator.memoize_new("$", set())


class ConstXlator (Xlator):
    """Translate all variable references with a constant string.

    Argument `subst` in `xlate` must be a string.
    """
    def _xlate(self, source, subst):
        assert isinstance(subst, six.string_types)
        return self._cre.sub(subst, str(source)), None

ConstXlator.memoize_new("@")
ConstXlator.memoize_new("@", None)
ConstXlator.memoize_new("@", set())
ConstXlator.memoize_new("$")
ConstXlator.memoize_new("$", None)
ConstXlator.memoize_new("$", set())


class ListXlator (Xlator):
    """Substitute the N'th variable with the N'th value in a sequence.

    Argument `subst` in `xlate` must be a list.
    """

    @functools.wraps(Xlator.__init__)
    def __init__(self, varre, ignore=None):
        self._tokenizer = Tokenizer(varre, ignore)

    def _xlate(self, source, subst):
        assert isinstance(subst, collections.Iterable) and \
            not isinstance(subst, six.string_types)
        res = ""
        idx = 0
        unxlated = []
        for tok in self._tokenizer.tokenize(source):
            if tok[0]:
                if idx >= len(subst):
                    unxlated.append(tok[1])
                else:
                    res += subst[idx]
                    idx += 1
            else:
                res += tok[1]
        if len(unxlated) is 0:
            unxlated = None
        return res, unxlated

ListXlator.memoize_new("@")
ListXlator.memoize_new("@", None)
ListXlator.memoize_new("@", set())
ListXlator.memoize_new("$")
ListXlator.memoize_new("$", None)
ListXlator.memoize_new("$", set())


class DictXlator (Xlator):
    """Substitute variable references with a dictionary mapping variable names
    to substitution values.

    Argument `subst` in `xlate` must be a dict-like object.
    """

    @functools.wraps(Xlator.__init__)
    def __init__(self, varre, ignore=None):
        Xlator.__init__(self, varre, ignore)
        self._tokenizer = Tokenizer(varre, ignore)

    def _xlate(self, source, subst):
        if not isinstance(subst, dict):
            raise TypeError("Tried to use a DictXlator "
                            "with a bad substitution type: '%s'" % subst)

        unxlated = []
        res = ""
        for isvar, tok in self._tokenizer.tokenize(source):
            if isvar:
                if tok in subst:
                    val = subst[tok]
                    if isinstance(val, six.string_types):
                        res += val
                    else:
                        res += str(val)
                else:
                    unxlated.append(tok)
                    res += self.var_start + tok + self.var_end
            else:
                res += tok
        if len(unxlated) is 0:
            unxlated = None
        return res, unxlated

DictXlator.memoize_new("@")
DictXlator.memoize_new("@", None)
DictXlator.memoize_new("@", set())
DictXlator.memoize_new("$")
DictXlator.memoize_new("$", None)
DictXlator.memoize_new("$", set())


class PrepDictXlator (DictXlator):
    """A `DictXlator` optimized for repeated substitutions on the same string.

    If argument `source` in `xlate` is ``None``, uses the source
    identified by a previous call to `prepare`. Otherwise, falls back to
    the `DictXlator` behaviour.

    Argument `subst` in `xlate` must be a dictionary.
    """
    @functools.wraps(Xlator.__init__)
    def __init__(self, varre, ignore=None):
        DictXlator.__init__(self, varre, ignore)
        self._prepared = None

    def prepare(self, source):
        """Prepare to start translating the given string."""
        self._prepared = self._cre.sub(r"%(\1)s", source)
        self._dict = dict(
            [(var, "%s%s%s" % (self.var_start, var, self.var_end))
             for var in self._tokenizer.vars(source)])
        self._keys = set(six.iterkeys(self._dict))

    def _xlate(self, source, subst):
        if source is not None:
            return DictXlator._xlate(self, source, subst)
        else:
            sdict = self._dict.copy()
            keys = set()
            for key, val in six.iteritems(subst):
                if key in sdict:
                    sdict[key] = val
                    keys.add(key)
            unxlated = self._keys - keys
            if len(unxlated) is 0:
                unxlated = None
            return self._prepared % sdict, unxlated
